# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

"""
Client for all `app` commands.
"""

from azdata.cli.commands.app.app_swagger import AppSwagger
from azdata.cli.commands.app.exceptions import (
    AppError,
    IllegalAppStateError
)
from azdata.cli.commands.app.util import cmd_inputs_to_dict
from azdata.cli.core.clients.cli_client import CliClient
from azdata.cli.core.exceptions import ApiError
from urllib.request import urlretrieve
from zipfile import ZipFile

import urllib
import time
import tempfile
import shutil
import os


def beget_app(_):
    return AppClientMixin()


class AppClientMixin(CliClient):
    def __init__(self):
        super(AppClientMixin, self).__init__()

    def swagger(self, name, version):
        try:
            res = self.request(
                '/api/v1/app/{0}/{1}/swagger.json'.format(name, version),
                'GET',
                response_type=object
            )

            return AppSwagger(res[0])
        except ApiError as e:
            raise AppError(e, status=e.status)
        except Exception as e:
            raise AppError(e)

    def run_app(self, name, version, inputs=None):
        try:
            res = self.apis.controller.list_app(name, version)

            if not res:
                raise ValueError('Application `{0}/{1}` not found.'.
                                 format(name, version))

            res = res[0]
            # -- enforce ready state (can run if it is not "Ready") --
            if not res.get('state') == 'Ready':
                raise IllegalAppStateError('Application `{0}/{1}` is not in a '
                                           '"Ready" state.'.
                                           format(name, version))

            input_defs = res.get('input_param_defs', [])
            app_params = cmd_inputs_to_dict(inputs, input_defs)
            api = self.import_api('app_api')
            swag = self.swagger(name, version)
            api.api_client.configuration.host = swag.host

            return api.run_app(
                name,
                version,
                app_params,
                swag.base_path,
                swag.path
            )
        except ApiError as e:
            raise AppError(e, status=e.status)
        except (ValueError, IllegalAppStateError) as e:
            raise AppError(e)
        except Exception as e:
            raise AppError(e)


# ============================================================================ #
# ============================================================================ #
# ============================================================================ #


def beget_tpl(kwargs):
    """Client factory for Template APIs."""
    from azdata.cli.commands.app.constants import TEMPLATE_URL

    url = kwargs['url'] if kwargs['url'] else TEMPLATE_URL
    return TemplateClientMixin(url)


class TemplateClientMixin(CliClient):
    """
    Client for the different Template APIS for the different supported runtime
    types.
    """

    GIT_SUFFIX = '.git'

    def __init__(self, url):
        super(TemplateClientMixin, self).__init__(check_auth=False)

        if url.endswith(self.GIT_SUFFIX):
            url = url.replace(self.GIT_SUFFIX, '/zipball/master/')
        self._url = url

    def list(self):
        """
        List all templates.
        """
        from collections import OrderedDict
        from azdata.cli.commands.app.spec import Spec

        stage_dir = None
        try:
            stage_dir = self.download()
            tpls = []

            for tpl in os.listdir(stage_dir):
                tpl_dir = os.path.join(stage_dir, tpl)
                if os.path.isdir(tpl_dir):
                    spec = Spec(tpl_dir)
                    o = OrderedDict()
                    o['template'] = tpl
                    o['runtime'] = spec.runtime
                    tpls.append(o)
            return tpls
        finally:
            shutil.rmtree(stage_dir)

    def download(self, name=None, dest=None):
        """
        Download application template(s).
        """
        import shutil

        archive = '.app-templates.zip'
        stage_dir = tempfile.mkdtemp()
        try:
            num_blocks = 0
            chunk_size = 4096
            req = urllib.request.urlopen(self._url)

            with open(archive, 'wb') as f:
                while True:
                    data = req.read(chunk_size)
                    time.sleep(.5)
                    num_blocks += 1

                    if not data:
                        break

                    f.write(data)

            # -- explode downloaded zip file --
            with ZipFile(archive, 'r') as zf:
                zf.extractall(stage_dir)

            # -- flatten zip content into a single dir and rename --
            content_dir = os.path.join(stage_dir, os.listdir(stage_dir)[0])

            if name:
                stage_dir = os.path.join(stage_dir, name)
                os.rename(os.path.join(content_dir, name), stage_dir)
            else:
                # all templates
                stage_dir = content_dir

            if dest:
                dest = os.path.join(dest, name) if name else dest
                shutil.move(stage_dir, dest)
            else:
                dest = stage_dir

            return dest
        finally:
            os.remove(archive)
