# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

"""
Spec

Overview
========

The start of declarative specifications for better workflow.

"""

from azdata.cli.commands.app.constants import RUNTIMES
from azdata.cli.commands.app.encoding_types import EncodingType
from azdata.cli.commands.app.exceptions import InvalidSpecError
from azdata.cli.core.logging import get_logger
from collections import OrderedDict

import yaml
import os
import six


log = get_logger(__name__)


class Spec(object):
    """
    Represents the application details as a YAML spec file (Read-Only).
    """

    FILE = 'spec.yaml'
    """
    The filename of the compliant YAML file.
    """

    SCHEMA = {
        'name': lambda x: type(x) == str,
        'version': lambda x: type(x) == str,
        'runtime': lambda x: type(x) == str,
        'src': lambda x: type(x) == str,
        'entrypoint': lambda x: type(x) == str,
        'owners': lambda x: type(x) == list,
        'readers': lambda x: type(x) == list,
        'replicas': lambda x: type(x) == int,
        'poolsize': lambda x: type(x) == int,
        'inputs': lambda x: type(x) == dict or type(x) == OrderedDict,
        'output': lambda x: type(x) == dict or type(x) == OrderedDict
    }
    """
    Defines the required shape of the YAML spec.
    """

    def __init__(self, app):
        # -- preserve yaml order --
        yaml.add_representer(
            OrderedDict,
            lambda dumper, data: dumper.represent_dict(six.iteritems(data))
        )
        yaml.add_constructor(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            lambda loader, node: OrderedDict(loader.construct_pairs(node))
        )

        if isinstance(app, dict):
            spec = OrderedDict(app)
        else:
            app = os.path.join(app, Spec.FILE) if os.path.isdir(app) else app
            with open(app) as f:
                spec = yaml.load(f, Loader=yaml.FullLoader)

        self._spec = spec
        self._validate()

    @property
    def name(self):
        """
        Gets the name.
        """
        return self._spec['name']

    @name.setter
    def name(self, name):
        """
        Sets the name.
        """
        self._spec['name'] = name

    @property
    def version(self):
        """
        Gets the version.
        """
        return self._spec['version']

    @version.setter
    def version(self, version):
        """
        Sets the version.
        """
        self._spec['version'] = version

    @property
    def runtime(self):
        """
        Gets the runtime.
        """
        return self._spec['runtime']

    def to_yaml(self):
        """
        Generates a spec.yaml as a str.
        """
        return yaml.dump(self._spec, default_flow_style=False)

    def dump(self, dest=os.getcwd()):
        """
        Writes this spec to a spec.yaml based on destination `dest`.
        """
        path = os.path.join(dest, Spec.FILE)
        with open(path, 'w') as f:
            yaml.dump(self._spec, f, default_flow_style=False)

        return path

    def tmp_yaml_file(self):
        """
        Writes this spec to a spec.yaml in a temporary location.
        """
        import tempfile

        tmp_dir = tempfile.mkdtemp()
        try:
            # -- for clean up, callers needs to manage --
            return {
                'directory': tmp_dir,
                'filename': Spec.FILE,
                'path': self.dump(tmp_dir)
            }
        except IOError as e:
            from azdata.cli.commands.app.util import silent_remove
            silent_remove([tmp_dir])
            raise IOError(e)

    def _validate(self):
        """
        Validate this `spec`.

        @NOTE: If spec yaml becomes more complicated, a true YAML schemata
        such as Rx should be adopted for validation.
        """
        from collections import Counter

        # -- should have valid runtime type --
        runtime = None
        if 'runtime' in self._spec:
            runtime = self._spec['runtime']

        if not runtime or runtime not in RUNTIMES:
            msg = 'Invalid runtime "{}" entity.'.format(runtime)
            raise InvalidSpecError(msg)

        # -- reasonable defaults if omitted --
        if 'replicas' not in self._spec:
            self._spec['replicas'] = 2

        if 'poolsize' not in self._spec:
            self._spec['poolsize'] = 2

        # TODO: replace validation here with server validation
        if runtime != 'R' and runtime != 'Python':
            return

        if 'owners' not in self._spec:
            self._spec['owners'] = []

        if 'readers' not in self._spec:
            self._spec['readers'] = []

        if 'inputs' not in self._spec:
            self._spec['inputs'] = OrderedDict()

        if 'output' not in self._spec:
            self._spec['output'] = OrderedDict()

        # - should have keys and valid types -
        for k, v in self._spec.items():
            if k in Spec.SCHEMA:
                if not Spec.SCHEMA[k](v):
                    msg = 'Invalid application spec entity type "{0}" for ' \
                          '"{1}".'.format(type(v).__name__, k)
                    raise InvalidSpecError(msg)
            else:
                msg = 'Invalid application spec entity "{}".'.format(k)
                raise InvalidSpecError(msg)

        # - should have identical keys in any order, no more no less --
        if Counter(list(self._spec.keys())) != \
                Counter(list(Spec.SCHEMA.keys())):
            msg = 'Invalid application spec entities.\n"{}"'.\
                format(list(self._spec.keys()))
            raise InvalidSpecError(msg)

        # -- should have valid io for runtime type and map client to server --
        try:
            inputs = {}
            for k, v in self._spec['inputs'].items():
                inputs[k] = EncodingType.cast_client_to_server_type(v, runtime)

            output = {}
            for k, v in self._spec['output'].items():
                output[k] = EncodingType.cast_client_to_server_type(v, runtime)

            self._spec['inputs'] = inputs
            self._spec['output'] = output
        except ValueError as e:
            raise InvalidSpecError(e)

        log.debug('The `spec.yaml` is valid.')
        log.debug(self.to_yaml())

