from azdata.cli.core.clients.cli_client import CliClient
from http import HTTPStatus
from azdata.cli.core.password import PasswordManager
from azdata.cli.core.deploy import display
from urllib3.exceptions import NewConnectionError, MaxRetryError

from azdata.cli.core.configuration import Configuration
from azdata.cli.core.clients.controller_client import ControllerClient
from azdata.cli.core.exceptions import ControllerError, http_status_codes
from azdata.cli.core.logging import get_logger
import azdata.cli.core.util as core_util
import azdata.cli.core.deploy as util
import azdata.cli.core.progress as progress_handler
import azdata.cli.core.kubernetes as kubernetes_util
import azdata.cli.core.upgrade as upgrade_util
import json
import os
import threading
import time

CONNECTION_RETRY_ATTEMPTS = 12
DELETE_CLUSTER_TIMEOUT_SECONDS = 300
RETRY_INTERVAL = 5
UPDATE_INTERVAL = (15 * 60) / RETRY_INTERVAL
logger = get_logger(__name__)


def beget(_):
    """Client factory"""
    return BdcClientMixin()


def beget_no_check_auth(_):
    """Client factory - no check on authentication"""
    return BdcClientMixin(False)


def beget_no_check_auth_or_eula(_):
    """Client factory - no check on authentication or eula"""
    return BdcClientMixin(False, False)


class BdcClientMixin(CliClient):
    def __init__(self, check_auth=True, check_eula=True):
        super(BdcClientMixin, self).__init__(check_auth, check_eula)
        self.cluster_name = None

    def create_bdc(self, cluster_name, config_object):
        """
        Create cluster
        :param cluster_name:
        :param config_object:
        :return:
        """
        # Set active context to None to ensure that if the user was previously
        # logged into a different cluster, they now need to re-login or
        # explicitly set their context and immediately persist to profile disk.
        self.profile.active_context = None
        self.profile.persist()

        self.cluster_name = cluster_name
        PasswordManager.check_password_validity("cluster")
        self.cluster_object = config_object
        self.cluster_secrets = dict()

        # Overwrite cluster name if it is specified on the command line.
        #
        if self.cluster_name is not None and self.cluster_name != "":
            self.cluster_object["metadata"]["name"] = self.cluster_name
        else:
            self.cluster_name = self.cluster_object["metadata"]["name"]

        # Keep track of status thread to join in case of deployment error
        stop = threading.Event()
        status_thread = threading.Thread(
            target=self.apis.control_deploy.get_controller_logs,
            args=[stop]
        )

        # After this stage the controller-client will be updated with an
        # endpoint, authenticated, and the authorization header set with a
        # `bearer access-token` intended for all future protected calls.
        self.apis.control_deploy.create_control_plane(
            self.cluster_object,
            self.cluster_name,
            status_thread
        )

        # Serialize the clusterObject back into the config
        #
        self.cluster_config = json.dumps(self.cluster_object)

        self.apis.controller.create_cluster(self.cluster_config)

        cluster_state = None
        i = 0
        while cluster_state != "ready":
            if cluster_state == "error":
                stop.set()
                status_thread.join()
                progress_handler.notify_deployment_error(self.cluster_name, cluster_state)
                raise Exception("Cluster '{0}' deployment failed. Cluster State: {1}."
                                .format(self.cluster_name, cluster_state))

            if i !=0 and i % UPDATE_INTERVAL == 0:
                logger.warning("Cluster '{0}' is not ready after {1} minutes. Check controller logs for more details."
                               .format(self.cluster_name, core_util.minutes_elapsed(i, RETRY_INTERVAL)))

            if cluster_state is not None:
                logger.debug("Waiting for cluster '{0}' to be in Ready state. Current state: {1}"
                             .format(self.cluster_name, cluster_state))

            time.sleep(RETRY_INTERVAL)
            cluster_state = self.apis.controller.get_cluster_state()
            i = i + 1

        # Poll to get remaining logs before notifying success
        #
        self.apis.control_deploy.get_controller_logs(stop, True)

        progress_handler.notify_deployment_success(self.cluster_name)

    def delete_bdc(self, cluster_name, force=False):
        """
        Delete a cluster
        """
        self.cluster_name = cluster_name
        display("Deleting cluster '%s'." % self.cluster_name)

        if not util.retry(kubernetes_util.namespace_exists, self.cluster_name,
                          retry_count=CONNECTION_RETRY_ATTEMPTS, retry_delay=RETRY_INTERVAL,
                          retry_method="check if namespace exists",
                          retry_on_exceptions=(NewConnectionError, MaxRetryError)):
            display("Cluster '%s' doesn't exist" % self.cluster_name)
            return

        # Check necessary environment variables for connecting to the controller.
        #
        cluster_state = None

        if self.profile.active_context.endpoint and self.profile.active_context.access_token \
                and self.profile.active_context.namespace == cluster_name:
            try:
                is_controller_ready = False
                for _ in range(2):
                    logger.debug("Checking if controller is ready to delete cluster '{0}'".format(self.cluster_name))
                    if self.apis.controller.is_ready(False):
                        is_controller_ready = True
                        break
                    time.sleep(RETRY_INTERVAL)
                if is_controller_ready:
                    cluster_state = self.apis.controller.get_cluster_state()
                    logger.debug("Deleting cluster '{0}' using the controller".format(self.cluster_name))

                    if cluster_state is not None:
                        self.apis.controller.delete_cluster()
                        cluster_state = self.apis.controller.get_cluster_state()

                    logger.debug("Waiting for cluster '{0}' to be deleted".format(self.cluster_name))
                    i = 1
                    while self is not None and cluster_state is not None:
                        time.sleep(RETRY_INTERVAL)
                        if i * RETRY_INTERVAL > DELETE_CLUSTER_TIMEOUT_SECONDS:
                            logger.warn("Cluster '{0}' is not deleted after {1} minutes."
                                        .format(self.cluster_name, DELETE_CLUSTER_TIMEOUT_SECONDS / 60))
                            break
                        if cluster_state != "Deleted":
                            logger.debug("Waiting for cluster '{0}' to be deleted. Current state : {1}."
                                         .format(self.cluster_name, cluster_state))

                        cluster_state = self.apis.controller.get_cluster_state(loggingEnabled=False)
                        i = i + 1
                else:
                    raise Exception("Failed to connect to controller")
            except ControllerError as e:
                # Raise the error if we encounter a 401 or 403 from controller
                # and do not fall back to kubernetes API for deletion.
                #
                if e.status in [http_status_codes.unauthorized, http_status_codes.forbidden]:
                    raise e
                else:
                    logger.debug(e)
            except Exception as e:
                logger.debug("Failed to delete the cluster '{0}' through controller.".format(self.cluster_name))
                logger.debug(e)
        elif not force:
            raise ValueError("Cluster name '{0}' does not match your azdata context. Please login to '{0}' with 'azdata"
                             " login' or force delete via your kube config with 'azdata bdc delete --name {0} --force'"
                             .format(cluster_name))

        # Try to delete the remaining resources in the cluster
        #
        i = 1
        resources_are_deleted = False
        http_status = None
        cluster_is_empty = False
        while not cluster_is_empty:

            time.sleep(RETRY_INTERVAL)

            if not resources_are_deleted:
                #  Try to delete the remaining resources in the cluster
                #
                (resources_are_deleted, http_status) = util.retry(kubernetes_util.delete_cluster_resources,
                                                                  self.cluster_name,
                                                                  retry_count=CONNECTION_RETRY_ATTEMPTS,
                                                                  retry_delay=RETRY_INTERVAL,
                                                                  retry_method="delete cluster resources",
                                                                  retry_on_exceptions=(NewConnectionError, MaxRetryError))

                if (http_status == HTTPStatus.FORBIDDEN):
                    break

            # Check if the cluster is empty
            #
            cluster_is_empty = util.retry(kubernetes_util.namespace_resources_are_deleted, self.cluster_name,
                                          retry_count=CONNECTION_RETRY_ATTEMPTS,
                                          retry_delay=RETRY_INTERVAL, retry_method="delete namespace resources",
                                          retry_on_exceptions=(NewConnectionError, MaxRetryError))

            if i * RETRY_INTERVAL > DELETE_CLUSTER_TIMEOUT_SECONDS:
                logger.warn("Cluster is not empty after %d minutes." % (DELETE_CLUSTER_TIMEOUT_SECONDS / 60))
                break

            i = i + 1
            time.sleep(RETRY_INTERVAL)

        if cluster_is_empty:
            display("Cluster '{0}' deleted successfully.".format(self.cluster_name))
        else:
            raise Exception("Failed to delete the cluster '{0}'.".format(self.cluster_name))

    def upgrade_bdc(self, cluster_name, target_version, target_repository=None):
        """
        Upgrade cluster
        :param cluster_name:
        :param target_version:
        :param target_repository:
        :return:
        """
        self.cluster_name = cluster_name

        kubernetes_util.validate_namespace(self.cluster_name)

        # Make sure the namespace exists
        #
        if not util.retry(kubernetes_util.namespace_exists, self.cluster_name, retry_count=CONNECTION_RETRY_ATTEMPTS,
                          retry_delay=RETRY_INTERVAL,
                          retry_method="check if namespace exists",
                          retry_on_exceptions=(NewConnectionError, MaxRetryError)):
            raise Exception("Upgrade cluster failed since the namespace %s does not exist." % self.cluster_name)

        if not self.apis.control_deploy.is_controller_pod_running(self.cluster_name):
            raise Exception(
                "Upgrade cluster failed since the controller is not deployed into namespace %s." % self.cluster_name)

        # Create the cluster patch
        #
        # TODO: Make this a proper JSON patch once config format is finalized
        #
        cluster_patch = {"targetVersion": target_version}

        # If the user specified a new repository then include that in the cluster patch as well
        #
        if target_repository:
            if "/" not in target_repository:
                raise Exception(
                    "Repository '%s' in not a valid value. The target respository should consist of the docker registry and \
                    docker repository separated by a forward slash ('/')." % target_repository)

            cluster_patch["targetRepository"] = target_repository

        # If the user passes in a new DOCKER_USERNAME and DOCKER_PASSWORD, update the docker registry secret
        #
        if ("DOCKER_USERNAME" in os.environ and os.environ.get("DOCKER_USERNAME")) and (
                "DOCKER_PASSWORD" in os.environ and os.environ.get("DOCKER_PASSWORD")):
            if not target_repository:
                    raise Exception("Target repository must be passed in order to update the docker credentials")

            docker_registry = target_repository.split("/")[0]

            util.retry(kubernetes_util.update_private_registry, cluster_name, docker_registry,
                       retry_count=CONNECTION_RETRY_ATTEMPTS,
                       retry_delay=RETRY_INTERVAL, retry_method="update docker private registry",
                       retry_on_exceptions=(NewConnectionError, MaxRetryError))

        display("Upgrading Control Plane.")

        # First upgrade the control plane
        #
        self.apis.controller.upgrade_control_plane(cluster_patch)

        # All control upgrade monitoring will be replaced by calls to the monitoring pod. Until then
        # we will manually check the upgrade config to see the status of the controller upgrade, and the
        # control plane state machine for the rest of the upgrade.
        #
        i = 0
        while upgrade_util.is_controller_upgrading(self.cluster_name):
            if i !=0 and i % UPDATE_INTERVAL == 0:
                logger.warning("Controller is still upgrading after %d minutes." % (core_util.minutes_elapsed(i, RETRY_INTERVAL)))

            time.sleep(RETRY_INTERVAL)
            i = i + 1

        if upgrade_util.check_controller_upgrade_failed(self.cluster_name):
            raise Exception("Control plane upgrade failed. Failed to upgrade controller.")

        # The controller and controller db have been upgraded. Now we need to wait for the rest service
        # to be reenabled
        #
        i = 0
        while not self.apis.controller.is_ready():
            time.sleep(RETRY_INTERVAL)
            i = i + 1


        control_plane_state = self.apis.controller.get_control_status(control_plane_state=True)

        # It can take a second or two for the state machine to move back into the upgrading state.
        # If the upgrade is still old, retry. We don't pause an upgrade until an image has been
        # updating for at least ten minutes, so we shouldn't get stuck in this loop.
        #
        while control_plane_state == "rollbackcomplete":
            time.sleep(RETRY_INTERVAL)

            control_plane_state = self.apis.controller.get_control_status(control_plane_state=True)

        stop = threading.Event()
        status_thread = threading.Thread(
            target=self.apis.control_deploy.get_controller_logs, args=[stop, False])
        status_thread.setDaemon(True)
        status_thread.start()

        i = 0
        while upgrade_util.is_control_plane_upgrading(self.cluster_name):

            if control_plane_state == "error":
                stop.set()
                status_thread.join()
                progress_handler.notify_control_plane_upgrade_error(self.cluster_name, control_plane_state)
                raise Exception("Control plane upgrade failed. Cluster State: %s." % control_plane_state)
            elif control_plane_state == "rollbackcomplete":
                progress_handler.notify_control_plane_upgrade_paused()
                return

            if i !=0 and i % UPDATE_INTERVAL == 0:
                logger.warning("Control plane is not ready after %d minutes." % core_util.minutes_elapsed(i, RETRY_INTERVAL))

            if control_plane_state is not None:
                logger.debug("Waiting for Control plane to be in Ready state. Current state : " + control_plane_state)

            time.sleep(RETRY_INTERVAL)
            control_plane_state = self.apis.controller.get_control_status(control_plane_state=True)
            i = i + 1

        progress_handler.notify_control_plane_upgrade_success()

        display("Upgrading Big Data Cluster.")

        # Upgrade the BDC
        #
        self.apis.controller.upgrade_bdc(cluster_patch)

        cluster_state = self.apis.controller.get_cluster_state()

        # It can take a second or two for the state machine to move back into the upgrading state.
        # If the upgrade is still old, retry. We don't pause an upgrade until an image has been
        # updating for at least ten minutes, so we shouldn't get stuck in this loop.
        #
        while cluster_state == "upgradepaused":
            time.sleep(RETRY_INTERVAL)

            cluster_state = self.apis.controller.get_cluster_state()

        i = 0
        while cluster_state != "ready":
            if cluster_state == "error":
                stop.set()
                status_thread.join()
                progress_handler.notify_bdc_upgrade_error(self.cluster_name, cluster_state)
                raise Exception("Cluster '{0}' upgrade failed. Cluster State: {1}"
                                .format(self.cluster_name, cluster_state))
            elif cluster_state == "upgradepaused":
                progress_handler.notify_bdc_upgrade_paused()
                return

            if i !=0 and i % UPDATE_INTERVAL == 0:
                logger.warning("Cluster '{0}' is not ready after {1} minutes."
                               .format(self.cluster_name, core_util.minutes_elapsed(i, RETRY_INTERVAL)))

            if cluster_state is not None:
                logger.debug("Waiting for cluster '{0}' to be in Ready state. Current state : {1}"
                             .format(cluster_name, cluster_state))

            time.sleep(RETRY_INTERVAL)
            cluster_state = self.apis.controller.get_cluster_state()
            i = i + 1

        progress_handler.notify_bdc_upgrade_success()
