# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

import os
import urllib3
import posixpath
from pywebhdfs import errors
from pywebhdfs.webhdfs import PyWebHdfsClient
from requests_kerberos import HTTPKerberosAuth
from azdata.cli.core.enums import SecurityStrategy
from azdata.cli.core.clients.cli_client import CliClient
from azdata.cli.core.exceptions import CliError
from azdata.cli.commands.hdfs.extwebhdfs import ExtendedPyWebHdfsClient
from azdata.cli.commands.hdfs.exceptions import HdfsError


# By default urllib3 prints a warning message to the console if using SSL and server certificate
# validation is not enabled.  This disables the warning.
urllib3.disable_warnings()


def beget(_):
    """Client factory"""
    return HdfsClientMixin(check_auth=True, check_eula=True)


class HdfsClientMixin(CliClient):
    block_size = 32 * (1 << 20)  # 32 MB

    def __init__(self, check_auth=True, check_eula=True):
        super(HdfsClientMixin, self).__init__(check_auth=check_auth, check_eula=check_eula)
        return

    def get_knox_endpoint(self):
        """
        Call the controller to get the webhdfs endpoint URL
        """
        try:
            endpoint = self.apis.controller.get_endpoints('webhdfs')['endpoint']

            # The endpoint must end with '/' in order for the APIs to succeed
            if not endpoint[len(endpoint) - 1] == '/':
                endpoint = endpoint + '/'
        except Exception as e:
            raise CliError(e)

        return endpoint

    def login(self, knox_url, username=None, password=None):
        """
        Login to the webhdfs endpoint
        """
        if self.profile.active_context.security_strategy == SecurityStrategy.basic:
            auth = (username, password)
        elif self.profile.active_context.security_strategy == SecurityStrategy.ad:
            auth = HTTPKerberosAuth()
        else:
            raise CliError("Invalid security strategy.  Use azdata login to choose supported strategy.")

        self.hdfs = ExtendedPyWebHdfsClient(
            base_uri_pattern=knox_url,
            request_extra_opts={'verify': False, 'auth': auth}
        )
        return

    def ls(self, path):
        """
        List status of the file or directory specified by the path.
        """
        result = self.hdfs.list_dir(path)
        return result

    def exists(self, path):
        """
        Determine if a file or directory exists.
        """
        return self.hdfs.exists_file_dir(path)

    def mkdir(self, path, permission = None):
        """
        Make a directory at the specified path.
        """
        if not permission == None:
            result = self.hdfs.make_dir(path, permission = permission)
        else:
            result = self.hdfs.make_dir(path)

        return result

    def mv(self, source_dir, target_dir):
        """
        Move a file from source to target.
        """
        return self.hdfs.rename_file_dir(source_dir, target_dir)

    def create_file(self, path, data = None):
        """
        create a file at the fiven path.
        """
        if not data == None:
            result = self.hdfs.create_file(path, file_data=data)
        else:
            result = self.hdfs.create_file(path, file_data=None)

        return result

    def cat_file(self, path, offset = 0, length = 4096):
        """
        Read the specified files content at the given offset and length.
        """
        return self.hdfs.read_file(path, offset = offset, length = length)

    def remove_file(self, path, recursive = False):
        """
        Remove the specified file.
        """
        return self.hdfs.delete_file_dir(path, recursive)

    def set_permission(self, path, permission):
        """
        Sets the permission on the specified file or directory.

        The permission needs to be a three octet string.  E.g. '775' or '755'.
        """
        return self.hdfs.set_permission(path, permission=permission)

    def set_owner(self, path, owner=None, group=None):
        """
        Sets the owner and group information for the given file or directory.
        """
        if owner == None:
            result = self.hdfs.set_owner(path, group=group)
        elif group == None:
            result = self.hdfs.set_owner(path, owner=owner)
        else:
            result = self.hdfs.set_owner(path, owner=owner, group=group)

        return result

    def modify_acl_entries(self, path, aclspec):
        """
        Add or modify ACL entries on the specified path.

        The aclspec is of the form '[user|group|other]:[name]:[r|-][w|-][x|-]'
        Multiple entries can be comma separated.
        Example: 'user:bob:rw-'
        """
        return self.hdfs.modify_acl_entries(path, aclspec=aclspec)

    def remove_acl_entries(self, path, aclspec):
        """
        Remove ACL entries on the specified path.

        The aclspec is of the form '[user|group|other]:[name]:[r|-][w|-][x|-]'
        Multiple entries can be comma separated.
        Example: 'user:bob:rw-'
        """
        return self.hdfs.remove_acl_entries(path, aclspec=aclspec)

    def remove_default_acl_entries(self, path):
        """
        Remove default ACL entries on the specified path: default:group, default:user, default:other
        """
        return self.hdfs.remove_default_acl_entries(path)

    def remove_acl(self, path):
        """
        Remove all but the base ACL entries.
        """
        return self.hdfs.remove_acl(path)

    def set_acl(self, path, aclspec):
        """
        Fully replaces the ACL entries with new values.

        The aclspec is of the form '[user|group|other]:[name]:[r|-][w|-][x|-]'
        Multiple entries can be comma separated.
        Example: 'user:bob:rw-'
        """
        return self.hdfs.set_acl(path, aclspec=aclspec)

    def get_acl_status(self, path):
        """
        Return the AclStatus JSON document for the given path.
        """
        return self.hdfs.get_acl_status(path)

    def get_xattr(self, path, name):
        """
        Gets the value of an extended attribute
        """
        return self.hdfs.get_xattr(path, xattr=name)

    def set_xattr(self, path, name, value):
        """
        Sets the value of an extended attribute
        """
        return self.hdfs.set_xattr(path, name, value, replace=False)

    def modify_xattr(self, path, name, value):
        """
        Modifies the value of an extended attribute
        """
        return self.hdfs.set_xattr(path, name, value, replace=True)

    def delete_xattr(self, path, name):
        """
        Modifies the value of an extended attribute
        """
        return self.hdfs.delete_xattr(path, name)

    def get_xattr_status(self, path):
        """
        Get all the extended attributes and their values for the given path.
        """
        return self.hdfs.get_xattr(path)

    def _copy_local_file_to_remote(self, root_path, local_path, remote_path, progress=None):
        """
        Copy a file page by page.  The local and remote paths must have been validated.
        """
        if not progress == None:
            # Get size of file
            file_size = os.path.getsize(local_path)
            progress.initialize(os.path.relpath(local_path, root_path), file_size)

        with open(local_path, 'rb') as f:
            # Create the remote file
            self.create_file(remote_path)

            done = False
            while not done:
                read_data = f.read(self.block_size)
                if len(read_data) > 0:
                    self.hdfs.append_file(remote_path, read_data)
                else:
                    done = True

                if not progress == None:
                    progress.update(len(read_data))

        return

    def _normalize_hdfs_path(self, path):
        # If the path contains back slashes replace them with forward slashes.
        #
        slash_convert = path.replace('\\','/')
        normalize_convert = posixpath.normpath(slash_convert)
        return normalize_convert


    def _copy_local_directory_to_remote(self, root_path, local_path, remote_path, progress=None):
        """
        Iterate through the given local directory copying each file to the remote location.
        The local and remote paths need to have been validated.
        """
        for dir_name, subdir_list, file_list in os.walk(local_path):
            local_relative_path = os.path.relpath(dir_name, local_path)
            remote_dir_path = self._normalize_hdfs_path(posixpath.abspath(posixpath.join(remote_path, local_relative_path)))
            for file_name in file_list:
                local_file = os.path.join(dir_name, file_name)
                remote_file = posixpath.join(remote_dir_path, file_name)
                self._copy_local_file_to_remote(root_path, local_file, remote_file, progress)
        return

    def _copy_local_to_remote(self, local_path, remote_path, progress=None):
        """
        Copy a local file or directory to the remote HDFS path.
        """
        if os.path.exists(local_path):
            if os.path.isdir(local_path):
                # The local path is a directory so the remote path must also be a directory.
                if not self.exists(remote_path):
                    self.mkdir(remote_path)
                    self._copy_local_directory_to_remote(local_path, local_path, remote_path, progress)
                else:
                    raise HdfsError("Remote directory [{0}] already exists.  Delete and then retry copy.".format(remote_path))
            elif os.path.isfile(local_path):
                # Check to see if the remote file exists.
                if not self.exists(remote_path):
                    self._copy_local_file_to_remote(os.path.dirname(local_path), local_path, remote_path, progress)
                else:
                    raise HdfsError("Remote file [{0}] already exists.  Delete and then retry copy.".format(remote_path))
            else:
                raise HdfsError("Local path [{0}] must be a normal file or directory.".format(local_path))
        else:
            raise FileNotFoundError(local_path)

        return True

    def _copy_remote_file_to_local(self, root_path, remote_path, local_path, progress=None):
        """
        Copy a file page by page.  The local and remote paths must have been validated.
        """
        if not progress == None:
            if not progress == None:
                # Get size of file
                stats = self.hdfs.get_file_dir_status(remote_path)
                file_size = stats["FileStatus"]["length"]
                progress.initialize(posixpath.relpath(remote_path, root_path), file_size)

        # Make sure the directory exists
        directory_path = os.path.dirname(local_path)
        if not os.path.exists(directory_path):
            os.makedirs(directory_path)

        with open(local_path, 'wb') as f:
            offset = 0

            done = False
            completed_size = 0
            while not done:
                try:
                    read_data = b''
                    read_data = self.hdfs.read_file(remote_path, offset=offset, length=self.block_size, buffersize=self.block_size)
                    if len(read_data) > 0:
                        f.write(read_data)
                    else:
                        done = True
                except errors.PyWebHdfsException as e:
                    # If the file being downloaded is a multiple of the block size then an exception will be thrown
                    # reading the block after end of file.  Most APIs simply return a zero size read but this API
                    # throws.  So catch the exception and check to make sure it's past end of file and exit.
                    if "out of the range".encode() in e.msg:
                        done = True
                    else:
                        raise e

                # REVIEW Need to confirm HDFS can't return short reads.  If you read an offset past the end
                # of the file a generic error is returned.  So no clean way to detect EOF except short read.
                #
                if len(read_data) < self.block_size:
                    done = True
                else:
                    offset += self.block_size

                if not progress == None:
                    progress.update(len(read_data))

        return

    def _copy_remote_directory_to_local(self, root_path, remote_path, local_path, progress=None):
        """
        Iterate through the given remote directory copying each file to the local location.
        The local and remote paths need to have been validated.
        """
        directory_contents = self.hdfs.list_dir(remote_path)
        remote_files = [x["pathSuffix"] for x in directory_contents["FileStatuses"]["FileStatus"] if x["type"] == "FILE"]
        remote_directories = [x["pathSuffix"] for x in directory_contents["FileStatuses"]["FileStatus"] if x["type"] == "DIRECTORY"]

        for remote_file in remote_files:
            self._copy_remote_file_to_local(root_path, posixpath.join(remote_path, remote_file), os.path.join(local_path, remote_file), progress)

        for remote_directory in remote_directories:
            self._copy_remote_directory_to_local(root_path, posixpath.join(remote_path, remote_directory), os.path.join(local_path, os.path.normpath(remote_directory)), progress)
        return

    def _copy_remote_to_local(self, remote_path, local_path, progress=None):
        """
        Copy a remote HDFS file or directory to the local path.
        """
        if self.exists(remote_path):
            remote_path_properties = self.hdfs.get_file_dir_status(remote_path)

            if remote_path_properties["FileStatus"]["type"] == "DIRECTORY":
                # Check to see if the local file exists.
                if not os.path.exists(local_path):
                    self._copy_remote_directory_to_local(remote_path, remote_path, local_path, progress)
                else:
                    raise HdfsError("Local directory [{0}] already exists.  Delete and then retry copy.".format(local_path))
            elif remote_path_properties["FileStatus"]["type"] == "FILE":
                # Check to see if the local file exists.
                if not os.path.exists(local_path):
                    self._copy_remote_file_to_local(posixpath.dirname(remote_path), remote_path, local_path, progress)
                else:
                    raise HdfsError("Local file [{0}] already exists.  Delete and then retry copy.".format(local_path))
            else:
                raise HdfsError("Remote file [{0}] is an unknown type.".format(remote_path))
        else:
            raise FileNotFoundError(remote_path)

        return True

    def _resolve_local_path_to_absolute(self, path):
        if os.path.isabs(path):
            return path
        else:
            abspath = os.path.join(os.getcwd(), path)
            return abspath

    def _resolve_hdfs_path(self, path):
        stripped_path = path.lstrip('hdfs:')
        if os.path.isabs(stripped_path):
            return stripped_path
        else:
            raise HdfsError("HDFS path must be absolute. [{0}]".format(stripped_path))

    def copy(self, from_path, to_path, progress):
        isFromHDFS = from_path.startswith('hdfs:')
        isToHDFS = to_path.startswith('hdfs:')

        if isFromHDFS and isToHDFS:
            raise HdfsError("Both paths are remote.  Remote to remote copy is not supported.  Copy requires one local and one remote path.")

        if not isFromHDFS and not isToHDFS:
            raise HdfsError("Both paths are local.  Local file copy is not supported.  Copy requires one local and one remote path.")

        if isToHDFS:
            return self._copy_local_to_remote(self._resolve_local_path_to_absolute(from_path), self._resolve_hdfs_path(to_path), progress=progress)
        else:
            return self._copy_remote_to_local(self._resolve_hdfs_path(from_path), self._resolve_local_path_to_absolute(to_path), progress=progress)
