import requests
from six.moves import http_client
from pywebhdfs.webhdfs import PyWebHdfsClient, _raise_pywebhdfs_exception

# Extended Operations
GETHOMEDIRECTORY = 'GETHOMEDIRECTORY'
SETPERMISSION = 'SETPERMISSION'
SETOWNER = 'SETOWNER'
MODIFYACLENTRIES = 'MODIFYACLENTRIES'
REMOVEACLENTRIES = 'REMOVEACLENTRIES'
REMOVEDEFAULTACL = 'REMOVEDEFAULTACL'
REMOVEACL = 'REMOVEACL'
SETACL = 'SETACL'
GETACLSTATUS = 'GETACLSTATUS'
CHECKACCESS = 'CHECKACCESS'
GETXATTRS = 'GETXATTRS'


class ExtendedPyWebHdfsClient(PyWebHdfsClient):
    """
    This class extends the PyWebHdfsClient with additional calls that the standard library doesn't
    implement.  By extending the class the same method implementation pattern can be used and the
    code reuse is maximized.

    Note there are no parts of the protocol implemented by this class that are different than
    the standard Apache Hadoop WebHdfs documentation.  No customization should eb done as part
    of this class.
    """

    def __init__(self, **kwargs):
        super(ExtendedPyWebHdfsClient, self).__init__(**kwargs)

    def get_home_directory(self):
        """
	    Gets the calling users home directory.

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/?op=GETHOMEDIRECTORY

        and returns a directory path.

        Example for getting home directory:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> hdfs.get_home_directory()
        {"Path": "/user/username"}

    	"""
        path = "/"
        response = self._resolve_host(requests.get, True,
                                      path, GETHOMEDIRECTORY)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return response.json()

    def set_permission(self, path, **kwargs):
        """
	    Sets the permission on a file or directory.

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=SETPERMISSION

        Example for setting permission:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.set_permission(my_file, '755')

        """
        optional_args = kwargs
        response = self._resolve_host(requests.put, True,
                                      path, SETPERMISSION, **optional_args)

        if response.status_code == http_client.OK:
            return True
        elif response.status_code == http_client.NOT_FOUND:
            return False
        _raise_pywebhdfs_exception(response.status_code, response.content)

    def set_owner(self, path, **kwargs):
        """
    	Sets the user and group owners for a file or directory

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=SETOWNER[&owner=<USER>][&group=<GROUP>]

        Example for setting user and group owner:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.set_owner(my_file, owner='hdfs', group='hdfs')

        """
        optional_args = kwargs
        response = self._resolve_host(requests.put, True,
                                      path, SETOWNER, **optional_args)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return True

    def modify_acl_entries(self, path, **kwargs):
        """
        Can add a new ACL entries or modify existing ACL entries.  Any ACL not specified is left unchanged.

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=MODIFYACLENTRIES&aclspec=<ACLSPEC>

        Example for modifying ACL entries:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.modify_acl_entries(my_file, aclspec='user:bob:rw-')

        """
        optional_args = kwargs
        response = self._resolve_host(requests.put, True,
                                      path, MODIFYACLENTRIES, **optional_args)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return True

    def remove_acl_entries(self, path, **kwargs):
        """
        Remove existing ACL entries.  Any ACL not specified is left unchanged.

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=REMOVEACLENTRIES&aclspec=<ACLSPEC>

        Note the aclspec doesn't have read/write/execute flags.

        Example for removing ACL entries:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.remove_acl_entries(my_file, aclspec='user:bob:')

        """
        optional_args = kwargs
        response = self._resolve_host(requests.put, True,
                                      path, REMOVEACLENTRIES, **optional_args)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return True

    def remove_default_acl_entries(self, path):
        """
        Remove default ACL entries: default:user, default:group, default:other

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=REMOVEDEFAULTACL

        Example for removing default ACL entries:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.remove_default_acl_entries(my_file)

        """
        response = self._resolve_host(requests.put, True,
                                      path, REMOVEDEFAULTACL)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return True

    def remove_acl(self, path):
        """
        Remove all but the base ACL entries.

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=REMOVEACL

        Example for removing an ACL:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.remove_acl(my_file)

        """
        response = self._resolve_host(requests.put, True,
                                      path, REMOVEACL)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return True

    def set_acl(self, path, **kwargs):
        """
        Fully replaces the ACL entries with new values.

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=SETACL&aclspec=<ACLSPEC>

        Example for setting an ACL:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.set_acl(my_file, aclspec='user:bob:rw-')

        """
        optional_args = kwargs
        response = self._resolve_host(requests.put, True,
                                      path, SETACL, **optional_args)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return True

    def get_acl_status(self, path):
        """
        Returns the AclStatus JSON object for the given path.

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=GETACLSTATUS

        Example for getting an ACL's status:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.get_acl_status(my_file)

        """
        response = self._resolve_host(requests.get, True,
                                      path, GETACLSTATUS)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return response.json()

    def get_xattr(self, path, xattr=None):
        """
        Get extended attributes set on an HDFS path.

        This function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=GETXATTRS

        Example for getting an extended attribute:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> hdfs.get_xattr('user/hdfs/data.txt', xattr='user.important')

        """
        kwd_params = {}
        if xattr:
            kwd_params['xattr.name'] = xattr
        kwd_params['encoding'] = 'text'

        response = self._resolve_host(requests.get, True,
                                      path, GETXATTRS,
                                      **kwd_params)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)
        return response.json()
