from azdata.cli.commands.hdfs.util import FormatWebHDFS
from azdata.cli.commands.hdfs import _
from pywebhdfs import errors

import os
import cmd
import shlex
import readline
import posixpath


class HDFSShell(cmd.Cmd):
    """
    This implements a simple interactive shell for navigating around a HDFS server.
    """
    intro = '\nHDFS Interactive Shell.  Type help or ? to list commands.\n'
    prompt = '[hdfs] '

    def __init__(self, client, knox_url, username=None, password=None):
        super().__init__()
        self.hdfs = client
        self.hdfs.login(knox_url, username, password)
        self.current_working_directory = "/"
        self.push_pop_stack = []
        readline.set_completer_delims(' \t\n')

    def _process_args(self, args):
        result = []
        lex = shlex.shlex(args, posix=True)
        lex.whitespace_split = True
        try:
            token = lex.get_token()
            while not token == lex.eof:
                result.append(token)
                token = lex.get_token()
        except ValueError:
            if not lex.token == None or not lex.token == '':
                result.append(lex.token)

        return result

    def _list_stats(self, path):
        result = None
        output = ""
        try:
            is_absolute = posixpath.isabs(path)
            if is_absolute:
                if self.hdfs.exists(path):
                    result = self.hdfs.ls(path)
                else:
                    output = "cd: No such file or directory."
            else:
                abs_path = posixpath.abspath(posixpath.join(self.current_working_directory, path))
                if self.hdfs.exists(abs_path):
                    result = self.hdfs.ls(abs_path)
                else:
                    output = "cd: No such file or directory."

            if result:
                converted_list = FormatWebHDFS.convert_filestatuses_to_list(result["FileStatuses"])
                output = FormatWebHDFS.format_directory_as_output(converted_list)

        except errors.FileNotFound as e:
            output = "ls: No such file or directory\n"

        return output

    def _complete_path_param(self, path):
        extra_text = None
        output = None

        # Scan the output.  If a path has quotes then remove them.
        if path.startswith('"') or path.startswith("'"):
            path = path[1:]
        if path.endswith('"') or path.endswith("'"):
            path = path[:len(path) - 1]

        # Resolve the path to an absolute path.
        is_absolute = posixpath.isabs(path)
        if not is_absolute:
            path = posixpath.abspath(posixpath.join(self.current_working_directory, path))

        # Check to see if what has been typed so far exists.  If not then
        # strip off the last part of the path.
        if not self.hdfs.exists(path):
            extra_text = posixpath.basename(path)
            path = posixpath.dirname(path)

        # Lookup the contents of the path
        if self.hdfs.exists(path if not path == "" else "/"):
            result = self.hdfs.ls(path if not path == "" else "/")
            if result:
                directory_list = FormatWebHDFS.convert_filestatuses_to_name_list(result["FileStatuses"])

                if not extra_text == None:
                    filtered_list = [d for d in directory_list if d.startswith(extra_text)]

                    if len(filtered_list) == 1:
                        output = [posixpath.join(path, filtered_list[0])]
                    else:
                        output = [posixpath.join(path, d) for d in filtered_list]
                else:
                    output = [posixpath.join(path, d) for d in directory_list]

        # Scan the output.  If a path has spaces then quote the value.
        for i in range(len(output)):
            if " " in output[i]:
                output[i] = '"{0}"'.format(output[i])

        if len(output) == 1:
            output[0] = output[0] + ' '

        return output

    # Commands
    def do_quit(self, arg):
        """Edit the interactive shell."""
        return True

    def do_ls(self, arg):
        """List status of a file or directory."""
        args = self._process_args(arg)

        # If multiple directories are provided show the content of all of them
        if len(args) == 0:
            output = self._list_stats(self.current_working_directory)
        elif len(args) > 1:
            output = ""
            for path in args:
                output += "\n{0}:\n".format(path)
                output += self._list_stats(path)
        else:
            output = self._list_stats(args[0])

        self.hdfs.stdout(output)

    def help_ls(self):
        output = shell_help_ls.format(
            short=_('List the status of the given file or directory. '),
            ex1=_('List Status'),
            ex2=_('List Status of a specific directory.'))

        self.hdfs.stdout(output)

    def complete_ls(self, text, line, begidx, endidx):
        args = self._process_args(line)
        if len(args) > 1:
            current_arg = args[len(args) - 1]
        else:
            current_arg = '/'
        output = self._complete_path_param(current_arg)
        return output

    def do_cd(self, arg):
        """
        Change the current directory to the speciied path.  The path can be
        relative or absolute.
        """
        args = self._process_args(arg)

        if len(args) > 0:
            # Note Ignore additional arguments and move to the first directory
            path = args[0]

            if path == "..":
                # Pop a directory off current
                abs_path = posixpath.abspath(posixpath.join(self.current_working_directory, os.pardir))
                if self.hdfs.exists(abs_path):
                    self.current_working_directory = abs_path
                    output = "cd: {0}".format(abs_path)
                else:
                    output = "cd: No such file or directory."
            else:
                is_absolute = posixpath.isabs(path)
                if is_absolute:
                    if self.hdfs.exists(path):
                        self.current_working_directory = path
                        output = "cd: {0}".format(path)
                    else:
                        output = "cd: No such file or directory."
                else:
                    abs_path = posixpath.abspath(posixpath.join(self.current_working_directory, path))
                    if self.hdfs.exists(abs_path):
                        self.current_working_directory = abs_path
                        output = "cd: {0}".format(abs_path)
                    else:
                        output = "cd: No such file or directory."

            self.hdfs.stdout(output)
        else:
            self.hdfs.stdout("cd: missing path argument.")

    def help_cd(self):
        output = shell_help_cd.format(
            short=_('Change the current directory to the speciied path.  The path can be relative or absolute.'),
            ex1=_('Change directory.'))

        self.hdfs.stdout(output)

    def do_pushd(self, arg):
        """
        Push the current directory onto the command stack and move to the
        requested directory.
        """
        args = self._process_args(arg)

        if len(args) > 0:
            # Note Ignore additional arguments and move to the first directory
            path = args[0]

            is_absolute = posixpath.isabs(path)
            if is_absolute:
                if self.hdfs.exists(path):
                    self.push_pop_stack.append(self.current_working_directory)
                    self.current_working_directory = path
                    output = "pushd: {0}".format(path)
                else:
                    output = "pushd: No such file or directory."
            else:
                abs_path = posixpath.abspath(posixpath.join(self.current_working_directory, path))
                if self.hdfs.exists(abs_path):
                    self.push_pop_stack.append(self.current_working_directory)
                    self.current_working_directory = abs_path
                    output = "pushd: {0}".format(abs_path)
                else:
                    output = "pushd: No such file or directory."

            self.hdfs.stdout(output)
        else:
            self.hdfs.stdout("pushd: missing path argument.")

    def help_pushd(self):
        output = shell_help_pushd.format(
            short=_('Push the current directory onto the command stack and move to the requested directory.'),
            ex1=_('Push directory.'))

        self.hdfs.stdout(output)

    def do_popd(self, arg):
        'Pop the last directory off the command stack and change directory to it.'
        if len(self.push_pop_stack) > 0:
            path = self.push_pop_stack.pop()
            self.do_cd(path)

    def help_popd(self):
        output = shell_help_popd.format(
            short=_('Pop the last directory off the command stack and change directory to it.'),
            ex1=_('Pop directory.'))

        self.hdfs.stdout(output)

    def do_mkdir(self, arg):
        'Make a new directory at the specified path.'
        args = self._process_args(arg)

        if len(args) > 0:
            # Note Ignore additional arguments and move to the first directory
            path = args[0]

            is_absolute = posixpath.isabs(path)
            if is_absolute:
                self.hdfs.mkdir(path)
            else:
                abs_path = posixpath.abspath(posixpath.join(self.current_working_directory, path))
                self.hdfs.mkdir(abs_path)
        else:
            self.hdfs.stdout("mkdir: missing path argument.")

    def help_mkdir(self):
        output = shell_help_mkdir.format(
            short=_('Create a directory at the specified path.'),
            ex1=_('Make directory.'))

        self.hdfs.stdout(output)

    def do_mv(self, arg):
        """Make a new directory at the specified path."""
        args = self._process_args(arg)

        if len(args) == 2:
            # Note Ignore additional arguments and move to the first directory
            from_dir = args[0]
            to_dir = args[1]

            if not posixpath.isabs(from_dir):
                from_dir = posixpath.abspath(posixpath.join(self.current_working_directory, from_dir))

            if not posixpath.isabs(to_dir):
                to_dir = posixpath.abspath(posixpath.join(self.current_working_directory, to_dir))

            if self.hdfs.exists(from_dir):
                self.hdfs.mv(from_dir, to_dir)
                output = "mv: Move completed"
            else:
                output = "mv: No such file or directory."
        else:
            output = "mv: Invalid arguments."

        self.hdfs.stdout(output)

    def help_mv(self):
        output = shell_help_mv.format(
            short=_('Move the specified file or path to the specified location.'),
            ex1=_('Move file or directory.'))

        self.hdfs.stdout(output)

    def do_create(self, arg):
        """
        Create a file if it doesn't exist.  Write the post path parameters to
        the file.
        """
        args = self._process_args(arg)

        if len(args) > 0:
            # Note all parameters after the path are written as the contents of the file.
            content = ""
            path = args[0]
            if len(args) > 1:
                content = " ".join(args[1:])

            is_absolute = posixpath.isabs(path)
            if not is_absolute:
                path = posixpath.abspath(posixpath.join(self.current_working_directory, path))

            if content is None:
                self.hdfs.create_file(path)
            else:
                self.hdfs.create_file(path, content)
        else:
            self.hdfs.stdout("create: missing path argument.")

    def help_create(self):
        output = shell_help_create.format(
            short=_(
                'Create the text file at the specified location.  Simple text content can be added via data parameter.'),
            ex1=_('Create file.'))

        self.hdfs.stdout(output)

    def do_cat(self, arg):
        """Make a new directory at the specified path."""
        args = self._process_args(arg)

        if len(args) > 0:
            # Note Ignore additional arguments and move to the first directory
            path = args[0]
            args = args[1:]

            offset = 0
            length = 4096
            i = 0
            while i < len(args):
                if args[i].lower() == "--offset":
                    offset = args[i + 1]
                    i += 2
                elif args[i].lower() == "--length":
                    length = args[i + 1]
                    i += 2
                else:
                    self.hdfs.stdout("cat: Error invalid argument [{0}]".format(args[i]))
                    return

            is_absolute = posixpath.isabs(path)
            if not is_absolute:
                path = posixpath.abspath(posixpath.join(self.current_working_directory, path))

            try:
                result = self.hdfs.cat_file(path, offset=offset, length=length)
                self.hdfs.stdout(result)
            except Exception as e:
                self.hdfs.stderr("cat: Error reading file. {0}".format(e))
        else:
            self.hdfs.stdout("cat: missing path argument.")

    def help_cat(self):
        output = shell_help_cat.format(
            short=_('Read content of a file.  Offset and length in bytes are optional parameters.'),
            ex1=_('Read file.'))

        self.hdfs.stdout(output)

    def do_rm(self, arg):
        """Remove a file or directory."""
        args = self._process_args(arg)

        if len(args) > 0:
            # Note Ignore additional arguments and move to the first directory
            path = args[0]

            is_absolute = posixpath.isabs(path)
            if not is_absolute:
                path = posixpath.abspath(posixpath.join(self.current_working_directory, path))

            try:
                self.hdfs.remove_file(path)
            except errors.FileNotFound:
                self.hdfs.stdout('rm: No such file or directory.')
        else:
            self.hdfs.stdout('rm: missing path argument.')

    def help_rm(self):
        output = shell_help_rm.format(
            short=_('Remove a file or directory.'),
            ex1=_('Remove a file or directory.'))

        self.hdfs.stdout(output)

    def do_rmr(self, arg):
        """Recursively remove a directory or file."""
        args = self._process_args(arg)

        if len(args) > 0:
            # Note Ignore additional arguments and move to the first directory
            path = args[0]

            is_absolute = posixpath.isabs(path)
            if not is_absolute:
                path = posixpath.abspath(posixpath.join(self.current_working_directory, path))

            try:
                self.hdfs.remove_file(path, recursive=True)
            except errors.FileNotFound as e:
                self.hdfs.stdout('rmr: No such file or directory.')
        else:
            self.hdfs.stdout("rmr: missing path argument.")

    def help_rmr(self):
        output = shell_help_rmr.format(
            short=_('Recursively remove a file or directory.'),
            ex1=_('Recursive remove directory.'))

        self.hdfs.stdout(output)

    def do_chmod(self, arg):
        'Change the permissions of the specified file or directory.'
        args = self._process_args(arg)

        if len(args) == 2:
            # Note Ignore additional arguments
            permission = args[0]
            path = args[1]

            if not posixpath.isabs(path):
                path = posixpath.abspath(posixpath.join(self.current_working_directory, path))

            if self.hdfs.exists(path):
                self.hdfs.set_permission(path, permission)
                output = "chmod: Change completed."
            else:
                output = "chmod: No such file or directory."
        else:
            output = "chmod: Invalid arguments."

        self.hdfs.stdout(output)

    def help_chmod(self):
        output = shell_help_chmod.format(
            short=_('Change the permission on the specified file or directory.'),
            ex1=_('Change file or directory permission.'))
        self.hdfs.stdout(output)

    def do_chown(self, arg):
        """Change the owner or group of the specified file or directory."""
        args = self._process_args(arg)

        if len(args) == 2:
            # Note Ignore additional arguments
            owner = args[0]
            group = None
            path = args[1]

            if not posixpath.isabs(path):
                path = posixpath.abspath(posixpath.join(self.current_working_directory, path))

            if ':' in owner:
                parts = owner.split(':')
                owner = parts[0]
                group = parts[1]

                owner = owner if not owner == "" else None
                group = group if not group == "" else None

            if self.hdfs.exists(path):
                self.hdfs.set_owner(path, owner=owner, group=group)
                output = "chown: Change completed."
            else:
                output = "chown: No such file or directory."
        else:
            output = "chown: Invalid arguments."

        self.hdfs.stdout(output)

    def help_chown(self):
        output = shell_help_chown.format(
            short=_('Change the owner or group of the specified file.'),
            ex1=_('Change the owner and group.'))
        self.hdfs.stdout(output)

    def do_acl(self, arg):
        """Change the owner or group of the specified file or directory."""
        args = self._process_args(arg)

        try:
            if len(args) < 2:
                raise ValueError('acl: Invalid arguments.')

            operation = args[0]
            path = args[1]

            if (operation == 'status' and not len(args) == 2) or \
                    (operation == 'remove' and
                        not (len(args) == 3 or len(args) == 3)) or \
                    (operation == 'set' and not len(args) == 3) or \
                    (operation == 'modify' and not len(args) == 3):

                raise ValueError('acl: Invalid arguments.')

            if operation == 'status':
                output = self.hdfs.get_acl_status(path)
            elif operation == 'remove':
                if len(args) == 2:
                    self.hdfs.remove_acl_entries(path)
                elif len(args) == 3 and args[2].lower() == 'default':
                    self.hdfs.remove_default_acl_entries(path)
                else:
                    aclspec = args[2]
                    self.hdfs.remove_acl_entries(path, aclspec)
                output = 'acl remove: Change completed.'
            elif operation == 'set':
                aclspec = args[2]
                self.hdfs.set_acl(path, aclspec)
                output = 'acl set: Change completed.'
            elif operation == 'modify':
                aclspec = args[2]
                self.hdfs.modify_acl_entries(path, aclspec)
                output = 'acl modify: Change completed.'
            else:
                raise ValueError('acl: Invalid command.  Must be one of '
                                 '[status, set, modify, remove]')

        except errors.FileNotFound:
            output = '{0}: No such file or directory'.format(operation)
        except (ValueError, Exception) as e:
            output = e

        self.hdfs.stdout(output)

    def help_acl(self):
        output = shell_help_acl.format(
            short=_("Advanced ACL related commands."),
            long=_("Commands for getting, setting, updating, and removing ACLs from files and directories.\n"
                   "These ACLs are in addition to POSIX ACLs.\n\n"
                   "The commands are:\n"
                   "  status = Get the AclStatus JSON document for the requested path.\n"
                   "  set    = Replace the ACL entries with the specified aclspec.\n"
                   "  modify = Update the ACL entries with the specified aclspec.\n"
                   "  remove = Remove the ACL entries in the specified aclspec.\n"
                   "           There are three aclspec values:\n"
                   "             None      = ACLs will be removed.\n"
                   "             'default' = Default ACLs will be removed.\n"
                   "             aclspec   = Specified entries will be removed.\n"),
            ex1=_("Get ACL status for a path."),
            ex2=_("Set ACL entries."),
            ex3=_("Modify ACL entries."),
            ex4=_("Remove ACL entries.")
        )
        self.hdfs.stdout(output)

    def do_xattr(self, arg):
        """Change the owner or group of the specified file or directory."""
        args = self._process_args(arg)

        if len(args) >= 2:
            operation = args[0]
            path = args[1]

            if operation == 'status':
                output = self.hdfs.get_xattr_status(path)
            elif len(args) == 3:
                name = args[2]
                if operation == 'get':
                    output = self.hdfs.get_xattr(path, name)
                elif operation == 'remove':
                    self.hdfs.delete_xattr(path, name)
                    output = "xattr remove: Change completed."
                else:
                    output = "xattr: Invalid arguments."
            elif len(args) == 4:
                name = args[2]
                value = args[3]
                if operation == 'set':
                    self.hdfs.set_xattr(path, name, value)
                    output = "xattr set: Change completed."
                elif operation == 'modify':
                    self.hdfs.modify_xattr(path, name, value)
                    output = "xattr modify: Change completed."
                else:
                    output = "xattr: Invalid arguments."
            else:
                output = "xattr: Invalid arguments."
        else:
            output = "xattr: Invalid arguments."

        self.hdfs.stdout(output)

    def help_xattr(self):
        output = shell_help_xattr.format(
            short=_("Extended attribute related commands."),
            long=_("Commands for getting, setting, updating, and removing extended attributes from files and directories.\n"
                   "The commands are:\n"
                   "  status = Get the extended attributes for the requested path.\n"
                   "  get    = Get the value for the given extended attribute.\n"
                   "  set    = Set the value for the extended attribute.  The attribute must not exist.\n"
                   "  modify = Modify the value for an extended attribute.  The attribute must exist.\n"
                   "  remove = Remove the extended attribute from the requested path.\n"),
            ex1=_("Get extended attribute status for a path."),
            ex2=_("Get extended attribute."),
            ex3=_("Set extended attribute."),
            ex4=_("Modify extended attribute."),
            ex5=_("Remove extended attribute.")
        )
        self.hdfs.stdout(output)

# pylint: disable=line-too-long
shell_help_ls = """
    type: command
    short-summary: {short}
    examples:
        - name: {ex1}
          text: >
            ls
        - name: {ex2}
          text: >
            ls '/tmp'
"""

# pylint: disable=line-too-long
shell_help_cd = """
    type: command
    short-summary: {short}
    examples:
        - name: {ex1}
          text: >
            cd '/tmp'
"""

# pylint: disable=line-too-long
shell_help_pushd = """
    type: command
    short-summary: {short}
    examples:
        - name: {ex1}
          text: >
            pushd '/tmp'
"""

# pylint: disable=line-too-long
shell_help_popd = """
    type: command
    short-summary: {short}
    examples:
        - name: {ex1}
          text: >
            popd
"""

# pylint: disable=line-too-long
shell_help_mkdir = """
    type: command
    short-summary: {short}
    examples:
        - name: {ex1}
          text: >
            mkdir '/tmp'
"""

# pylint: disable=line-too-long
shell_help_mv = """
    type: command
    short-summary: {short}
    examples:
        - name: {ex1}
          text: >
            mv '/tmp' '/dest'
"""

# pylint: disable=line-too-long
shell_help_create = """
    type: command
    short-summary: {short}
    examples:
        - name: {ex1}
          text: >
            create '/tmp/test.txt' "This is a test."
"""

# pylint: disable=line-too-long
shell_help_cat = """
    type: command
    short-summary: {short}
    examples:
        - name: {ex1}
          text: >
            cat '/tmp/test.txt'
"""

# pylint: disable=line-too-long
shell_help_rm = """
    type: command
    short-summary: {short}
    examples:
        - name: {ex1}
          text: >
            rm '/tmp'
"""

# pylint: disable=line-too-long
shell_help_rmr = """
    type: command
    short-summary: {short}
    examples:
        - name: {ex1}
          text: >
            rmr '/tmp'
"""

# pylint: disable=line-too-long
shell_help_chmod = """
    type: command
    short-summary: {short}
    examples:
        - name: {ex1}
          text: >
            chmod 775 '/tmp/test.txt'
"""

# pylint: disable=line-too-long
shell_help_chown = """
    type: command
    short-summary: {short}
    examples:
        - name: {ex1}
          text: >
            chown hdfs:superusergroup '/tmp/test.txt'
"""

# pylint: disable=line-too-long
shell_help_acl = """
    type: command
    short-summary: {short}
    long-summary: {long}
    examples:
        - name: {ex1}
          text: >
            acl status '/tmp/test.txt'
        - name: {ex2}
          text: >
            acl set '/tmp/test.txt' 'user:bob:r--'
        - name: {ex3}
          text: >
            acl modify '/tmp/test.txt' 'user:bob:rwx'
        - name: {ex4}
          text: >
            acl remove '/tmp/test.txt' 'user:bob:'
"""

# pylint: disable=line-too-long
shell_help_xattr = """
    type: command
    short-summary: {short}
    long-summary: {long}
    examples:
        - name: {ex1}
          text: >
            xattr status '/tmp/test.txt'
        - name: {ex2}
          text: >
            xattr get '/tmp/test.txt' 'myprop'
        - name: {ex3}
          text: >
            xattr set '/tmp/test.txt' 'myprop' '1234'
        - name: {ex4}
          text: >
            xattr modify '/tmp/test.txt' 'myprop' '5678'
        - name: {ex5}
          text: >
            xattr remove '/tmp/test.txt' 'myprop'
"""
