# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from azdata.cli.core.enums import ContextType
from azdata.cli.core.clients.cli_client import CliClient
from azdata.cli.core.profile import (
    CliProfile,
    select_authenticator
)


def beget(_):
    """Client factory for user's authentication profile."""
    return ProfileClientMixin()


class ProfileClientMixin(CliClient):
    """
    User profile client.
    """

    def __init__(self):
        super(ProfileClientMixin, self).__init__(
            check_auth=False, check_eula=False)

        self.active_context = self.profile.active_context

    def set_profile(
            self,
            accept_eula,
            endpoint,
            username,
            security_strategy,
            token,
            context_type):

        # -- persist --
        CliProfile(
            accept_eula=accept_eula,
            endpoint=endpoint,
            username=username,
            security_strategy=security_strategy,
            auth_token=token,
            context_type=context_type
        )

    def login(self, args):
        authenticator = select_authenticator(args)
        token = authenticator.authenticate()

        # -- persist token and endpoint into the profile --
        self.set_profile(
            args.get('accept_eula', 'no'),
            authenticator.endpoint,
            authenticator.username,
            args.get('auth'),
            token,
            ContextType.bdc.name
        )

        return token.namespace

    def logout(self):
        try:
            profile = CliProfile()
            if profile.active_context.namespace:
                self.delete_context(profile.active_context.namespace)
        except Exception as e:
            raise e

    def endpoint(self):
        return self.active_context.endpoint

    def accept_eula(self):
        return self.profile.accept_eula

    def namespace(self):
        return self.active_context.namespace

    def list_context(self, active=False):
        try:
            if not active:
                contexts = CliProfile().contexts
                if contexts:
                    for key, value in contexts.items():
                        contexts[key]['accessToken'] = '<PRIVATE>'
                    return contexts
                else:
                    return None
            else:
                if CliProfile().active_context.namespace:
                    context = CliProfile().active_context.serialize()
                    context['accessToken'] = '<PRIVATE>'
                    return context
                else:
                    return None
        except Exception as e:
            raise ValueError('Failed to list contexts - {0}'.format(e))

    def delete_context(self, name):
        try:
            profile = CliProfile()
            contexts = profile.contexts
            if contexts:
                if name in contexts:
                    del contexts[name]
                else:
                    raise ValueError('Context with namespace {0} does not '
                                     'exist'.format(name))

                if name == profile.active_context.namespace:
                    profile.active_context = None
                if contexts:
                    profile.contexts = contexts
                else:
                    profile.contexts = None
            else:
                raise ValueError('No contexts exist.  Please login to a '
                                 'control plane.')
        except Exception as e:
            raise ValueError('Failed to delete {0} - {1}'.format(name, e))

    def set_context(self, name):
        try:
            profile = CliProfile()
            if profile.contexts:
                if name in profile.contexts:
                    profile.active_context = profile.contexts[name]
                else:
                    raise ValueError('Context with namespace {0} does not '
                                     'exist'.format(name))
            else:
                raise ValueError('No contexts exist.  Please login to a '
                                 'control plane.')
        except Exception as e:
            raise ValueError('Failed to set context as {0} - {1}'.
                             format(name, e))