# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from knack.prompting import NoTTYException
from azdata.cli.core.configuration import Configuration
from azdata.cli.core.enums import SecurityStrategy
from azdata.cli.core.prompt import prompt
from azdata.cli.core.exceptions import (
    ApiError,
    CliError,
    KerberosExchangeError,
    HTTP_UNAUTHORIZED
)
from azdata.cli.core.util import (
    load_kube_config,
    get_controller_endpoint,
    raise_command_error, eula_prompt
)

import os


def login(
        client,
        auth=None,
        endpoint=None,
        accept_eula=None,
        namespace=None,
        # --Basic --
        username=None,
        # -- AD --
        principal=None):

    """
    ```sh
    azdata login
    ```
    """
    try:
        # get a flexible map of arguments, we do this to support different
        # authentication types in the future
        args = locals()

        # -- default to basic authentication --
        if args['auth'] is None:
            args['auth'] = SecurityStrategy.basic.name

        if username and args['auth'] == SecurityStrategy.ad.name:
            raise ValueError('Argument "--username" can not be used with the '
                             '"--auth ad" security strategy.')

        if principal and args['auth'] == SecurityStrategy.basic.name:
            raise ValueError('Argument "--principle" can not be used with the '
                             '"--auth basic" security strategy.')

        # -- Acquire/normalize EULA --
        args['accept_eula'] = eula_prompt(client, accept_eula, True).lower()

        # -- Acquire/normalize endpoint --
        if not endpoint:
            # 1. identify namespace
            if not namespace:
                try:
                    namespace = prompt('Namespace: ')
                except NoTTYException:
                    raise NoTTYException('When not specifying the --endpoint '
                                         'and using the kube config, please '
                                         'specify the namespace as an '
                                         'argument via --namespace/-n.')
                args['namespace'] = namespace

            # 2. Attempt to load the kube config if it exists to get endpoint
            if os.environ.get('KUBECONFIG', None) or \
                    os.environ.get('KUBECTL_CONTEXT', None) \
                    or os.path.isfile(os.path.expanduser('~/.kube/config')):
                try:
                    # Attempt to get endpoint via kube config
                    load_kube_config(os.getenv('KUBECTL_CONTEXT', None))
                    endpoint = get_controller_endpoint(namespace)

                except Exception as e:
                    raise CliError('Please check your kube config or specify '
                                   'the correct controller endpoint with: '
                                   '--controller-endpoint https://<ip>:<port>. '
                                   '{0}'.format(e))

            # 3. Check Profile to get endpoint
            endpoint = client.endpoint() if not endpoint else endpoint

            # 4. Prompt to get endpoint
            if not endpoint:
                try:
                    endpoint = prompt('Endpoint: ')
                except NoTTYException:
                    raise NoTTYException('Missing --endpoint arg in '
                                         'non-interactive environment. You may '
                                         'also place your kube config in the '
                                         'default location of <user home>/'
                                         '.kube/config or use the KUBECONFIG '
                                         'env var to specify the location of '
                                         'the config, i.e. export KUBECONFIG='
                                         'path/to/.kube/config.')
            args['endpoint'] = endpoint

        args['endpoint'] = args['endpoint'].rstrip('/')

        # -- login with dedicated authenticator based on arguments --
        official_namespace = client.login(args)

        msg = 'Logged in successfully to `{0}` in namespace `{1}`. Setting ' \
              'active context to `{1}`.'.format(endpoint, official_namespace)

        if namespace is not None and namespace != official_namespace:
            msg += '\nThere is a mismatch between the namespace and endpoint ' \
                   'you specified, using the endpoint\'s namespace `{0}`.'.\
                format(official_namespace)

        client.stdout(msg)

    except ValueError as ex:
        raise CliError(ex)
    except KerberosExchangeError as ex:
        raise CliError('No credentials were supplied, or the credentials were '
                       'unavailable or inaccessible.')
    except ApiError as ex:
        if ex.status == HTTP_UNAUTHORIZED:
            raise CliError('Wrong username or password.')
        else:
            raise_command_error(ex)
    except NoTTYException as e:
        raise CliError('{0} Please specify arguments and/or environment '
                       'variables in non-interactive mode. For a better '
                       'understanding of the arguments, please run \'{1} '
                       'login --help\'.'.format(Configuration().CLI_NAME, e))
    except Exception as ex:
        raise CliError(ex)


def logout(client):
    """
    ```sh
    azdata logout
    ```
    """
    try:
        client.logout()
    except Exception as e:
        raise CliError(e)


def context_set(client, name):
    """
    ```sh
    azdata context set
    ```
    """
    try:
        return client.set_context(name)
    except Exception as e:
        raise CliError(e)


def context_list(client, active=False):
    """
    ```sh
    azdata context list
    ```
    """
    try:
        return client.list_context(active)
    except Exception as e:
        raise CliError(e)


def context_delete(client, name):
    """
    ```sh
    azdata context delete
    ```
    """
    try:
        client.delete_context(name)
    except Exception as e:
        raise CliError(e)