# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from azdata.cli.core.clients.cli_client import CliClient
from mssqlcli.mssqlclioptionsparser import create_parser
from mssqlcli.main import run_cli_with
from mssqlcli.config import get_config
from azdata.cli.core.exceptions import CliError
from azdata.cli.core import Configuration
from configobj import ConfigObj
from os.path import expanduser, isfile
from azdata.cli.core.enums import SecurityStrategy

import os
import pyodbc


def beget(_):
    """Client factory"""
    return SqlClientMixin()


class SqlClientMixin(CliClient):
    def __init__(self, check_auth=True, check_eula=True):
        super(SqlClientMixin, self).__init__(check_auth, check_eula)

        # Add config file option value
        config_path = self.get_config_path()

        if not isfile(config_path):
            # The CLI has it's own configuration file.  With the integration it
            # made sense to have a separate configuration so users can configure
            # them differently if they want.  Since the file is identical they
            # can also copy over the file to make them the same.  This imports
            # the CLI code and uses that to create the default config file then
            # edits it to not spew info output to the console.
            get_config(config_path)
            cfg = ConfigObj(
                expanduser(config_path),
                interpolation=False,
                encoding='utf-8'
            )
            cfg.filename = expanduser(config_path)
            cfg["main"]["log_level"] = "NONE"
            cfg.write()

    def get_config_path(self):
        profile_path = os.path.expanduser(Configuration().CLI_CONFIG_DIR)
        config_path = os.path.join(profile_path, "mssql-cli.config")
        return config_path

    def get_cluster_endpoint(self, endpoint_name=None):
        try:
            return self.apis.controller.get_endpoints(endpoint_name)['endpoint']
        except Exception as e:
            raise CliError(e)

    def run_shell(self, endpoint, username, password):
        security_strategy = self.profile.active_context.security_strategy

        if security_strategy == SecurityStrategy.basic:
            args = [
                "-S", endpoint,
                "-U", username,
                "-P", password
            ]
        elif security_strategy == SecurityStrategy.ad:
            args = [
                "-S", endpoint,
                "-E"
            ]
        else:
            raise CliError("Invalid security strategy.  Use azdata login to "
                           "choose supported strategy.")

        os.environ["MSSQL_CLI_TELEMETRY_OPTOUT"] = "1"
        mssqlcli_options_parser = create_parser()
        mssqlcli_options = mssqlcli_options_parser.parse_args(args)

        config_path = self.get_config_path()
        mssqlcli_options.mssqlclirc_file = config_path
        run_cli_with(mssqlcli_options)

        return

    def run_query(self, endpoint, username, password, database, query):
        connection_info = endpoint.split(',')
        if len(connection_info) == 1:
            host = connection_info[0]
            port = 1433
        else:
            host = connection_info[0]
            port = connection_info[1]

        security_strategy = self.profile.active_context.security_strategy

        if security_strategy == SecurityStrategy.basic:
            conn = pyodbc.connect(
                driver='{ODBC Driver 17 for SQL Server}',
                server='{0},{1}'.format(host, port),
                database=database,
                uid=username,
                pwd=password,
                autocommit=True
            )
        elif security_strategy == SecurityStrategy.ad:
            conn = pyodbc.connect(
                driver='{ODBC Driver 17 for SQL Server}',
                server='{0},{1}'.format(host, port),
                database=database,
                trusted_connection='yes',
                autocommit=True
            )
        else:
            raise CliError("Invalid security strategy.  Use azdata login to "
                           "choose supported strategy.")

        cursor = conn.cursor()
        cursor.execute(query)

        # pyodbc can return some empty result sets that end up throwing
        # ProgrammingErrror.  Skip these
        #
        done = False
        cursor_results = None
        while not done:
            try:
                cursor_results = cursor.fetchall()
                done = True
            except pyodbc.ProgrammingError:
                available = cursor.nextset()
                if not available:
                    done = True

        if cursor_results:
            result = []
            column_names = []
            for column_index in range(len(cursor_results[0])):
                column_metadata_name = cursor.description[column_index][0]
                if column_metadata_name:
                    column_names.append(column_metadata_name)
                else:
                    column_names.append('Column{0}'.format(column_index))

            for row_id in range(len(cursor_results)):
                row = {}
                for column_id in range(len(cursor_results[0])):
                    row[column_names[column_id]] = cursor_results[row_id][column_id]

                result.append(row)
        else:
            result = []

        conn.close()

        return result
