# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

import os
from azdata.cli.core.util import raise_command_error
from azdata.cli.core.prompt import prompt_pass
from knack.prompting import NoTTYException
from azdata.cli.core.enums import SecurityStrategy
from azdata.cli.core.exceptions import CliError
from azdata.cli.core.constants import (AZDATA_USERNAME, AZDATA_PASSWORD)


def _get_sql_username(client):
    if not os.getenv(AZDATA_USERNAME):
        try:
            client.stdout(
                "The SQL Server username is needed.  It can be set for the session by "
                "setting the environment variable {}.".format(AZDATA_USERNAME))
            username = prompt_pass('Enter SQL Server username: ')
        except NoTTYException:
            raise NoTTYException('Missing {} environment variable in non-interactive environment.'
                                 .format(AZDATA_USERNAME))
    else:
        username = os.environ[AZDATA_USERNAME]

    return username

def _get_sql_password(client):
    # Note for now if the environment variable isn't set prompt for the password..
    if not os.getenv(AZDATA_PASSWORD):
        try:
            client.stdout(
                "The SQL Server password is needed.  It can be set for the session by "
                "setting the environment variable {}.".format(AZDATA_PASSWORD))
            password = prompt_pass('Enter SQL Server Password: ')
        except NoTTYException:
            raise NoTTYException('Missing {} environment variable in non-interactive environment.'
                                 .format(AZDATA_PASSWORD))
    else:
        password = os.environ[AZDATA_PASSWORD]

    return password


def _get_login_data(client):
    endpoint = client.get_cluster_endpoint("sql-server-master")
    if client.profile.active_context.security_strategy == SecurityStrategy.basic:
        username = _get_sql_username(client)
        password = _get_sql_password(client)
    elif client.profile.active_context.security_strategy == SecurityStrategy.ad:
        username = None
        password = None
    else:
        raise CliError("Invalid security strategy.  Use azdata login to choose supported strategy.")

    return endpoint, username, password


def sql_shell(client):
    try:
        endpoint, username, password = _get_login_data(client)
        result = client.run_shell(endpoint, username, password)
    except Exception as ex:
        raise_command_error(ex)


def sql_query(client, database, query):
    try:
        endpoint, username, password = _get_login_data(client)
        result = client.run_query(endpoint, username, password, database, query)
        return result
    except Exception as ex:
        raise_command_error(ex)

