# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

"""
Client for all CLI actions.
"""

from azdata.cli.core.output import OutputStream
from azdata.cli.core.profile import CliProfile
from azdata.cli.core.clients.process_client import Command as ProcessClient
from azdata.cli.core.clients.controller_client import ControllerClient
from azdata.cli.core.clients.control_deploy_client import ControlDeployClient
from azdata.cli.core.clients.kerberos_client import KerberosClient
from azdata.cli.core.exceptions import CliError
from azdata.cli.core.configuration import Configuration
from azdata.cli.core.util import eula_prompt
from abc import ABCMeta
from six import add_metaclass

__all__ = ['client']


def beget_cli_client(check_auth=True, check_eula=True):
    """
    The factory function used to apply the common `CliClient` to a custom
    commands's command group.

    :param check_auth: Check for authentication on this command group.
    :param check_eula: Check for eula acceptance on this command group.
    :return: A function.
    """
    def beget(_):
        return CliClient(check_auth=check_auth, check_eula=check_eula)

    return beget

client = beget_cli_client  # Export

# ============================================================================ #
# ============================================================================ #
# ============================================================================ #


@add_metaclass(ABCMeta)
class BaseCliClient(object):
    def __init__(self, check_auth=True, check_eula=True):
        self._profile = CliProfile()
        if check_eula:
            accept_eula = self.profile.accept_eula
            if not accept_eula:
                self.profile.accept_eula = eula_prompt(self)

        if check_auth:
            if not self.profile.active_context or \
                    not self.profile.active_context.endpoint or \
                    not self.profile.active_context.access_token:
                err = 'Please run \'{0} login\' to first authenticate, or ' \
                      'set a valid active context with \'{0} context set\'.'.\
                    format(Configuration().CLI_NAME)
                raise CliError(err)

    @property
    def profile(self):
        return self._profile

    @property
    def stdout(self):
        return OutputStream().stdout.write

    @property
    def stderr(self):
        return OutputStream().stderr.write

# ============================================================================ #
# ============================================================================ #
# ============================================================================ #


class CliClient(BaseCliClient):
    """
    Default client injected in every command group. For further command
    customization extend this class.
    """
    def __init__(self, check_auth=True, check_eula=True):
        super(CliClient, self).__init__(check_auth, check_eula)

        ac = self.profile.active_context

        # -- exposed client APIS --
        # follow the same pattern when more "clients" need to be introduced
        self._apis = type('', (object,), {
            'process': ProcessClient,
            'controller': ControllerClient(ac.endpoint, ac.access_token),
            'control_deploy': ControlDeployClient(self, ac.namespace),
            'kerberos': KerberosClient(ac.endpoint, principal=ac.username)
        })

    @property
    def apis(self):
        """
        Gets the reference to the different API resource clients.
        """
        return self._apis

    @property
    def request(self):
        """
        Utility to acquire a low-level request function to preform general HTTP
        operations.

        :return: A request function.
        """

        def _request(resource_path,
                     method,
                     path_params=None,
                     query_params=None,
                     header_params=None,
                     body=None,
                     post_params=None,
                     files=None,
                     response_type=None,
                     auth_settings=None,
                     return_http_data_only=None,
                     collection_formats=None,
                     preload_content=True,
                     request_timeout=None):
            """
            Makes the HTTP request (synchronous) and returns deserialized data.

            :param resource_path: Path to method endpoint.
            :param method: Method to call.
            :param path_params: Path parameters in the url.
            :param query_params: Query parameters in the url.
            :param header_params: Parameters to be placed in the request header.
            :param body: body: Request body.
            :param post_params: Request post form parameters, for
                                `application/x-www-form-urlencoded`,
                                `multipart/form-data`.
            :param files: key -> filename, value -> filepath, for
                          `multipart/form-data`.
            :param response_type: Response data type.
            :param auth_settings: Auth Settings names for the request.
            :param return_http_data_only: response data without head status code
                                          and headers
            :param collection_formats: dict of collection formats for path,
            query,
                                       header, and post parameters.
            :param preload_content: if False, the urllib3.HTTPResponse object
            will
                                    be returned without reading/decoding
                                    response
                                    data. Default is True.
            :param request_timeout: timeout setting for this request. If one
                                    number provided, it will be total request
                                    timeout. It can also be a pair (tuple) of
                                    (connection, read) timeouts.
            :return: The response directly.
            """

            return self.apis.controller.api_client.call_api(
                resource_path,
                method,
                path_params=path_params,
                query_params=query_params,
                header_params=header_params,
                body=body,
                post_params=post_params,
                files=files,
                response_type=response_type if response_type else object,
                auth_settings=auth_settings if auth_settings else [],
                _return_http_data_only=return_http_data_only,
                collection_formats=collection_formats,
                _preload_content=preload_content,
                _request_timeout=request_timeout
            )

        return _request

    def import_api(self, api_module):
        """
        Gets the instance of the API class from the given `api_module`. This is
        a common pattern for instantiated APIs given a API module name.

        :param api_module: The name of API module.
        :return: The instance of the API.
        """
        from importlib import import_module
        from inspect import (getmembers, isclass)

        mod = import_module('azdata.cli.core.api.{}'.format(api_module))
        name, klass = getmembers(mod, isclass)[1]
        return klass(api_client=self.apis.controller.api_client)
