from kubernetes import client as k8sClient
from kubernetes.client.rest import ApiException as K8sApiException
from azdata.cli.core.profile import acquire_token_by_basic
from azdata.cli.core.password import PasswordManager
from azdata.cli.core.configuration import Configuration
from azdata.cli.core.clients.controller_client import ControllerClient
from azdata.cli.core.deploy import display, is_active_directory_security_enabled
from urllib3.exceptions import NewConnectionError, MaxRetryError

from azdata.cli.core.logging import get_logger
from azdata.cli.core.util import get_controller_endpoint
import azdata.cli.core.deploy as util
import azdata.cli.core.progress as progress_handler
import azdata.cli.core.kubernetes as kubernetes_util
import azdata.cli.core.upgrade as upgrade_util
import base64
import getpass
import json
import os
import sys
import threading
import time
import yaml

CONNECTION_RETRY_ATTEMPTS = 12
RETRY_INTERVAL = 5
SCRIPT_PATH = os.path.dirname(os.path.realpath(__file__))
logger = get_logger(__name__)


class ControlDeployClient(object):
    """
    The client for managing creation of a control plane
    """
    def __init__(self, cli_client, namespace=None):
        self.cli_client = cli_client
        self.namespace = namespace
        self.logs_lock = threading.Lock()
        self.offset = 0

    def create_control_plane(self, config, namespace=None, status_thread=None):
        """
        Create control plane
        :param config:
        :param namespace:
        :return:
        """
        PasswordManager.check_password_validity("cluster")
        cluster_secrets = dict()

        # Overwrite cluster name if it is specified on the command line.
        #
        if namespace is not None and namespace != "":
            config["metadata"]["name"] = namespace
        else:
            namespace = config["metadata"]["name"]

        self.namespace = namespace

        # Setup Private Docker Registry
        #
        docker_registry = \
            config["spec"]["controlPlane"]["spec"]["docker"]["registry"]

        # Create Kubernetes namespace
        #
        self.create_namespace(docker_registry, config, self.namespace)

        cluster_secrets["clusterName"] = self.namespace

        self.create_controller(config, cluster_secrets)

        controller_endpoint = get_controller_endpoint(self.namespace)
        ip_endpoint = get_controller_endpoint(self.namespace, True)

        if controller_endpoint == ip_endpoint:
            endpoint_str = controller_endpoint
        else:
            endpoint_str = controller_endpoint + ', ' + ip_endpoint

        progress_handler.notify_service_endpoint_available("controller", endpoint_str)
        
        # Use ip endpoint for basic auth during deployment
        controller_endpoint = ip_endpoint

        # Get the initial throw away controller client based on basic auth for
        # the initial bootstrapping. Once controller has been stood up we ignore
        config = Configuration()
        config.username = os.environ['AZDATA_USERNAME']
        config.password = os.environ['AZDATA_PASSWORD']
        tmp_controller_client = ControllerClient(controller_endpoint)
        tmp_controller_client.set_authorization_header(
            'Basic',
            config.get_basic_auth_token(strip_prefix=True)
        )

        i = 0
        while not tmp_controller_client.is_ready():
            # Log to console once every 5 minutes if controller is not ready
            if i != 0 and i % 60 == 0:
                display("Waiting for control plane to be ready after %d "
                        "minutes." % ((i * RETRY_INTERVAL) / 60))

            time.sleep(RETRY_INTERVAL)
            i = i + 1

        # Update the controller client ref with the endpoint and with an
        # authenticated access-token such that all future calls are allowed
        # access to protected resources via Bearer auth going forward in this
        # process.
        token = acquire_token_by_basic(
            controller_endpoint,
            username=config.username,
            password=config.password
        )

        self.cli_client.apis.controller.endpoint = controller_endpoint
        self.cli_client.apis.controller.set_authorization_header(
            'Bearer',
            token.access_token
        )

        # Serialize the clusterObject back into the config
        #
        if not status_thread:
            stop = threading.Event()
            status_thread = threading.Thread(
                target=self.get_controller_logs, args=[stop])

        status_thread.setDaemon(True)
        status_thread.start()

    def create_controller(
            self,
            config,
            secrets,
            node_admin_login='nodeadmin',
            knox_username='root'):
        """
        Create controller
        :param config: Control config to be deployed
        :param secrets: Secrets of the control
        :param node_admin_login:
        :param knox_username:
        :return:
        """
        if self.is_controller_pod_running(self.namespace):
            display("Controller already exists.")
            return

        # Check necessary environment variables for controller.
        #
        util.check_environment_variables("controller")

        # Check if passed in passwords pass the complexity requirements.
        #
        PasswordManager.check_password_validity("controller")

        # Create controller database SA password
        #
        controller_sa_password = PasswordManager.generate_random_sql_linux_password()
        secrets["controllerSaPassword"] = controller_sa_password

        # Create admin login password
        #
        node_admin_password = PasswordManager.generate_random_sql_linux_password()

        timestamp = str(int(time.time()))

        # Save base64 values in the cluster secrets object
        #
        if is_active_directory_security_enabled(config["spec"]["controlPlane"]):
            secrets["base64DomainServiceAccountPassword"] = base64.b64encode(
                bytes(os.environ["DOMAIN_SERVICE_ACCOUNT_PASSWORD"], 'utf-8')).decode('utf-8')
            secrets["base64DomainServiceAccountUsername"] = base64.b64encode(
                bytes(os.environ["DOMAIN_SERVICE_ACCOUNT_USERNAME"], 'utf-8')).decode('utf-8')
        else:
            secrets["base64DomainServiceAccountPassword"] = ""
            secrets["base64DomainServiceAccountUsername"] = ""

        secrets["base64ControllerUsername"] = base64.b64encode(
            bytes(os.environ["AZDATA_USERNAME"], 'utf-8')).decode('utf-8')
        secrets["base64ControllerPassword"] = base64.b64encode(
            bytes(os.environ["AZDATA_PASSWORD"], 'utf-8')).decode('utf-8')
        secrets["base64ControllerSaPassword"] = base64.b64encode(
            bytes(controller_sa_password, 'utf-8')).decode('utf-8')
        secrets["base64KnoxUsername"] = base64.b64encode(
            bytes(knox_username, 'utf-8')).decode('utf-8')
        secrets["base64KnoxPassword"] = base64.b64encode(
            bytes(os.environ["AZDATA_PASSWORD"], 'utf-8')).decode('utf-8')
        secrets["base64NodeAdminLogin"] = base64.b64encode(
            bytes(node_admin_login, 'utf-8')).decode('utf-8')
        secrets["base64NodeAdminPassword"] = base64.b64encode(
            bytes(node_admin_password, 'utf-8')).decode('utf-8')
        secrets["base64Timestamp"] = base64.b64encode(
            bytes(timestamp, 'utf-8')).decode('utf-8')
        secrets["base64mssqlSaPassword"] = base64.b64encode(
            bytes(os.environ["AZDATA_PASSWORD"], 'utf-8')).decode('utf-8')

        configs = {
            "controllerReplicaSetConfig": "rs-controller.yaml.tmpl",
            "controllerSvcConfig": "svc-controller.yaml.tmpl",
            "controllerEndpointSvcConfig": "svc-controller-external.yaml.tmpl",
            "namespacedAdminRoleConfig": "role-namespaced-admin.yaml",
            "controllerConfigMapConfig": "cm-controller.json.tmpl",
            "controllerDatabaseStatefulSetConfig": "ss-controldb.yaml.tmpl",
            "controllerDatabaseSvcConfig": "svc-controldb.yaml.tmpl"
        }

        secret_configs = {
            "domainServiceAccountSecretConfig": "secret-domain-service-account.yaml.tmpl",
            "controllerLoginSecretConfig": "secret-controller-login.yaml.tmpl",
            "controllerSaSecretConfig": "secret-controller-sa.yaml.tmpl",
            "nodeAdminLoginSecretConfig": "secret-node-admin-login.yaml.tmpl",
            "mssqlSaSecretConfig": "secret-mssql-sa.yaml.tmpl"
        }

        if "image" not in config["spec"]["controlPlane"]["spec"]["docker"]:
            config["spec"]["controlPlane"]["spec"]["docker"]["image"] = "mssql-controller"

        # Update deployment templates
        #
        for key, value in configs.items():
            template_config = util.get_config_from_template(os.path.join(SCRIPT_PATH, "../templates", value), config)
            configs[key] = template_config

        # Update secret templates
        #
        for key, value in secret_configs.items():
            template_config = util.get_config_from_template(os.path.join(SCRIPT_PATH, "../templates", value), secrets)
            secret_configs[key] = template_config

        display("Starting cluster deployment.")

        # Create secrets for controller
        #
        for _, value in secret_configs.items():
            util.retry(kubernetes_util.create_secret, self.namespace, value,
                       retry_count=CONNECTION_RETRY_ATTEMPTS,
                       retry_delay=RETRY_INTERVAL, retry_method="create secrets",
                       retry_on_exceptions=(NewConnectionError, MaxRetryError))

        # Create empty agent secrets for controller and controller db
        # Controller will modify them with the right data
        #
        util.retry(kubernetes_util.create_empty_secret, self.namespace, "controldb-agent-secret",
                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                   retry_delay=RETRY_INTERVAL, retry_method="create agent secret for controller db",
                   retry_on_exceptions=(NewConnectionError, MaxRetryError))

        util.retry(kubernetes_util.create_empty_secret, self.namespace, "control-agent-secret",
                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                   retry_delay=RETRY_INTERVAL, retry_method="create agent secret for controller",
                   retry_on_exceptions=(NewConnectionError, MaxRetryError))

        # Create controller service account, namespaced role and role binding
        #
        try:
            if not util.retry(kubernetes_util.service_account_exists, self.namespace, "sa-mssql-controller",
                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                   retry_delay=RETRY_INTERVAL, retry_method="check if service account exists",
                   retry_on_exceptions=(NewConnectionError, MaxRetryError)):

                   body = k8sClient.V1ServiceAccount()
                   body.metadata = k8sClient.V1ObjectMeta(name="sa-mssql-controller",
                                                          labels={'MSSQL_CLUSTER': self.namespace})
                   k8sClient.CoreV1Api().create_namespaced_service_account(namespace=self.namespace, body=body)

            if not util.retry(kubernetes_util.namespaced_role_exists, self.namespace, "namespaced-admin",
                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                   retry_delay=RETRY_INTERVAL, retry_method="check if role exists",
                   retry_on_exceptions=(NewConnectionError, MaxRetryError)):

                   body = yaml.full_load(configs["namespacedAdminRoleConfig"])
                   k8sClient.RbacAuthorizationV1Api().create_namespaced_role(namespace=self.namespace, body=body)

            role_binding_name = "rb-" + self.namespace + "-controller"
            if not util.retry(kubernetes_util.namespaced_role_binding_exists, self.namespace, role_binding_name,
                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                   retry_delay=RETRY_INTERVAL, retry_method="check if role binding exists",
                   retry_on_exceptions=(NewConnectionError, MaxRetryError)):

                   body = k8sClient.V1RoleBinding(
                       metadata=k8sClient.V1ObjectMeta(name=role_binding_name,
                                                       namespace=self.namespace,
                                                       labels={'MSSQL_CLUSTER': self.namespace}),
                       subjects=[
                            k8sClient.V1Subject(kind="ServiceAccount", name="sa-mssql-controller",
                                                namespace=self.namespace)],
                       role_ref=k8sClient.V1RoleRef(name="namespaced-admin", kind="Role",
                                                    api_group="rbac.authorization.k8s.io"))
                   k8sClient.RbacAuthorizationV1Api().create_namespaced_role_binding(namespace=self.namespace, body=body)
        except K8sApiException as e:
            logger.error(e.body)
            raise

        # Create controller pod and service
        #
        try:
            body = json.loads(configs["controllerConfigMapConfig"])

            # Add control plane config to the config map
            #
            body["data"]["mssql-controller.conf"] = json.dumps(config["spec"]["controlPlane"])
            body["metadata"]["labels"] = {'MSSQL_CLUSTER': self.namespace}

            node_label = None
            cluster_label = None

            if "clusterLabel" in config["spec"]["controlPlane"]["spec"]:
                cluster_label = config["spec"]["controlPlane"]["spec"]["clusterLabel"]

            if "nodeLabel" in config["spec"]["controlPlane"]["spec"]:
                node_label = config["spec"]["controlPlane"]["spec"]["nodeLabel"]

            util.add_node_selector_to_controller_config_map(body, cluster_label, node_label)

            # Create controller ConfigMap
            #
            k8sClient.CoreV1Api().create_namespaced_config_map(namespace=self.namespace, body=body)

            # Create Cluster ConfigMap
            #
            data = {}
            kubernetes_util.create_config_map(cluster_name=self.namespace, config_map_name="cluster-configmap",
                                              data=data)

            # Create PVCs if persistent storage is defined.
            #
            if "storage" in config["spec"]["controlPlane"]["spec"]:
                control_pods = ["controller", "controldb"]

                # Create PVs for logs and data for the controller and controller db pods.
                #
                for pod in control_pods:
                    pvc_config = util.get_config_from_template(
                        os.path.join(SCRIPT_PATH, "../templates", "pvc-logs-control.yaml.tmpl"), config)
                    body = yaml.full_load(pvc_config)
                    body["metadata"]["name"] = "logs-%s" % pod
                    body["metadata"]["labels"] = {'MSSQL_CLUSTER': self.namespace}
                    k8sClient.CoreV1Api().create_namespaced_persistent_volume_claim(namespace=self.namespace,
                                                                                    body=body)

                    pvc_config = util.get_config_from_template(
                        os.path.join(SCRIPT_PATH, "../templates", "pvc-data-control.yaml.tmpl"), config)
                    body = yaml.full_load(pvc_config)
                    body["metadata"]["name"] = "data-%s" % pod
                    body["metadata"]["labels"] = {'MSSQL_CLUSTER': self.namespace}
                    k8sClient.CoreV1Api().create_namespaced_persistent_volume_claim(namespace=self.namespace,
                                                                                    body=body)

            # Deploy controller db stateful set
            #
            body = yaml.full_load(configs["controllerDatabaseStatefulSetConfig"])
            util.add_node_selector_to_controller_config(body, cluster_label, node_label)

            k8sClient.AppsV1Api().create_namespaced_stateful_set(namespace=self.namespace, body=body)

            # Deploy controller replica set
            #
            body = yaml.full_load(configs["controllerReplicaSetConfig"])
            util.add_node_selector_to_controller_config(body, cluster_label, node_label)

            k8sClient.AppsV1Api().create_namespaced_replica_set(namespace=self.namespace, body=body)

            # Create controller service
            #
            body = yaml.full_load(configs["controllerSvcConfig"])
            k8sClient.CoreV1Api().create_namespaced_service(namespace=self.namespace, body=body)

            # Create controller db service
            #
            body = yaml.full_load(configs["controllerDatabaseSvcConfig"])
            k8sClient.CoreV1Api().create_namespaced_service(namespace=self.namespace, body=body)

            # Create controller endpoint service
            #
            body = yaml.full_load(configs["controllerEndpointSvcConfig"])
            k8sClient.CoreV1Api().create_namespaced_service(namespace=self.namespace, body=body)
        except K8sApiException as e:
            logger.error(e.body)
            raise

        wait_time = 0

        while not self.is_controller_pod_running(self.namespace):
            time.sleep(RETRY_INTERVAL)
            wait_time = wait_time + RETRY_INTERVAL

            if wait_time % 30 == 0:
                display("Waiting for cluster controller to start.")

    def create_namespace(self, docker_registry, config, namespace):
        """
        Create kubernetes namespace
        :param docker_registry:
        :param config:
        :param namespace:
        :return:
        """
        kubernetes_util.validate_namespace(namespace)

        # Initialize and check storage. Ensure storage classes exist for persistent volumes.
        #
        util.retry(self.initialize_and_check_storage,
                   config,
                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                   retry_delay=RETRY_INTERVAL,
                   retry_method="list k8s storage classes",
                   retry_on_exceptions=(NewConnectionError, MaxRetryError))

        # Create namespaces for the new cluster
        #
        if util.retry(kubernetes_util.namespace_exists, namespace, retry_count=CONNECTION_RETRY_ATTEMPTS,
                      retry_delay=RETRY_INTERVAL,
                      retry_method="check if namespace exists",
                      retry_on_exceptions=(NewConnectionError, MaxRetryError)):
            if not util.retry(kubernetes_util.namespace_is_empty, namespace,
                          retry_count=CONNECTION_RETRY_ATTEMPTS, retry_delay=RETRY_INTERVAL,
                          retry_method="check if namespace is empty",
                          retry_on_exceptions=(NewConnectionError, MaxRetryError)):
                raise Exception(
                    "Cluster creation not initiated because the existing namespace %s is not empty. Run \"kubectl get all -n %s \""
                    " to see the existing objects in the namespace" %
                    (namespace, namespace))
        else:
            util.retry(kubernetes_util.create_namespace, namespace,
                       retry_count=CONNECTION_RETRY_ATTEMPTS,
                       retry_delay=RETRY_INTERVAL, retry_method="create namespace",
                       retry_on_exceptions=(NewConnectionError, MaxRetryError))

        if ("DOCKER_USERNAME" in os.environ and os.environ.get("DOCKER_USERNAME")) and (
                "DOCKER_PASSWORD" in os.environ and os.environ.get("DOCKER_PASSWORD")):
            util.retry(kubernetes_util.setup_private_registry, namespace, docker_registry,
                       retry_count=CONNECTION_RETRY_ATTEMPTS,
                       retry_delay=RETRY_INTERVAL, retry_method="set up docker private registry",
                       retry_on_exceptions=(NewConnectionError, MaxRetryError))

    @staticmethod
    def is_controller_pod_running(namespace):
        """
        Check if controller is running.
        :return:
        """
        try:
            controller_service = k8sClient.CoreV1Api().read_namespaced_service(name="controller-svc-external",
                                                                               namespace=namespace)
            if controller_service.spec.type == 'LoadBalancer':
                if controller_service.status.load_balancer.ingress is None:
                    logger.debug(
                        "Waiting for public IP to be assigned for the controller service, please check the kubernetes "
                        "events for more details.")
                    return False

            # check if replicas are available for controller
            #
            rs_output = k8sClient.AppsV1Api().read_namespaced_replica_set(name="control", namespace=namespace)
            if rs_output.status.available_replicas != rs_output.status.replicas:
                logger.debug(
                    "Controller pod(s) are not in running state yet, please check the kubernetes events for more "
                    "details.")
                return False

            return True

        except K8sApiException as e:
            logger.debug(e.body)
            return False

    def get_controller_logs(self, stop, execute_once=False):
        """
        Read Controller logs
        :param stop: threading stop event, when triggered it will stop the infinite loop
        :param execute_once:
        :return:
        """
        try:
            while not stop.is_set():
                try:
                    self.logs_lock.acquire()
                    response = self.cli_client.apis.controller.\
                        get_controller_logs(self.offset)
                    progress_handler.log_progress(response)
                    if not response:
                        pass
                    elif not response.endswith('\r\n'):
                        self.offset += len(response.rsplit('\r\n', 1)[0])
                        continue
                    else:
                        self.offset += len(response)
                finally:
                    self.logs_lock.release()
                if execute_once:
                    return
                time.sleep(RETRY_INTERVAL)
        except Exception as e:
            display(e)

    def set_password(self, service_name, knox_username):
        """
        Updates password for the given user
        :param service_name:
        :param knox_username:
        :return:
        """

        # Check necessary environment variables for controller.
        #
        util.check_environment_variables("password")

        # Validate cluster name
        #
        if self.namespace.lower() == "default":
            raise Exception("Cluster name can not be \"default\".")

        # Validate that the controller is running and get its endpoint
        #
        if not self.is_controller_pod_running(self.namespace):
            raise Exception("Controller is not running.")

        # Read username and the new password
        #
        if service_name == "controller":
            password_var = "CONTROLLER_NEW_PASSWORD"
        else:
            raise Exception(
                "Unsupported service for set-password: '%s'. Valid services are {controller}." % service_name)

        if password_var in os.environ:
            new_password = os.environ[password_var]
        else:
            new_password = getpass.getpass("Enter new password: ")
            if new_password != getpass.getpass("Re-enter new password: "):
                raise Exception("The passwords do not match.")

        # Run password complexity check
        #
        if not PasswordManager.validate_password_and_print_if_error(new_password):
            logger.warning("Password complexity check failed. Exiting...")
            sys.exit(1)

        # Issue password update
        #
        if service_name == "controller":
            self.set_password_through_controller(service_name, new_password, knox_username)
        else:
            raise Exception(
                "Unsupported service for set-password: '%s'. Valid services are: {controller}." % service_name)

    @staticmethod
    def initialize_and_check_storage(config):
        """
        Check if storage class exists.
        :param config:
        :return:
        """
        storage_class_list = k8sClient.StorageV1Api().list_storage_class().items
        storage_class_names = [s.metadata.name for s in storage_class_list]

        control_plane = config["spec"]["controlPlane"]
        if "storage" in control_plane["spec"]:
            storage = control_plane["spec"]["storage"]

            if "data" not in storage:
                config["spec"]["controlPlane"]["spec"]["storage"]["data"] = {}
                config["spec"]["controlPlane"]["spec"]["storage"]["data"]["className"] = storage[
                    "className"]
                config["spec"]["controlPlane"]["spec"]["storage"]["data"]["accessMode"] = storage[
                    "accessMode"]
                config["spec"]["controlPlane"]["spec"]["storage"]["data"]["size"] = storage["size"]

            if "logs" not in storage:
                config["spec"]["controlPlane"]["spec"]["storage"]["logs"] = {}
                config["spec"]["controlPlane"]["spec"]["storage"]["logs"]["className"] = storage[
                    "className"]
                config["spec"]["controlPlane"]["spec"]["storage"]["logs"]["accessMode"] = storage[
                    "accessMode"]
                config["spec"]["controlPlane"]["spec"]["storage"]["logs"]["size"] = storage["size"]

            storage = config["spec"]["controlPlane"]["spec"]["storage"]

            if storage["data"]["className"] not in storage_class_names:
                raise Exception(storage["data"]["className"] + " storage class does not exist")
            if storage["logs"]["className"] not in storage_class_names:
                raise Exception(storage["logs"]["className"] + " storage class does not exist")
