# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

import knack.commands


def _load_module(loader, name):
    from importlib import import_module
    prefix = 'azdata.cli.commands.'
    module = import_module(prefix + name)
    loader_cls = getattr(module, 'COMMAND_LOADER_CLS', None)
    return loader_cls(cli_ctx=loader.cli_ctx)


def _load_module_command_loader(loader, args, name):
    command = ''
    for arg in args:
        if arg.startswith('-'):
            break
        command = '{} {}'.format(command, arg).strip()

    command_loader = _load_module(loader, name)
    command_table = command_loader.load_command_table(args)

    if command in command_table:
        loader._loaders_to_cmd[command] = command_loader

    return command_table


class CliCommandsLoader(knack.commands.CLICommandsLoader):
    """CLI command loader, containing the command mappings."""

    def __init__(self, *args, **kwargs):
        self._custom_command = kwargs.get('custom_command_type')
        self._loaders_to_cmd = {}
        super(CliCommandsLoader, self).__init__(
            *args,
            excluded_command_handler_args=['client'],
            **kwargs)

    def load_command_table(self, args):
        """Load all CLI commands by grouping."""

        from importlib import import_module
        import pkgutil
        import traceback

        try:
            mods_ns_pkg = import_module('azdata.cli.commands')
            installed_command_modules = [modname for _, modname, _ in
                                         pkgutil.iter_modules(
                                             mods_ns_pkg.__path__)]
        except ImportError:
            pass

        for mod in installed_command_modules:
            merge = self.command_table.copy()
            merge.update(_load_module_command_loader(self, args, mod))
            self.command_table = merge

        return self.command_table

    def load_arguments(self, command):
        """Load all CLI arguments for command."""

        loader = self._loaders_to_cmd.get(command, None)
        if loader and command in self.command_table:
            loader.command_name = command
            self.command_table[command].load_arguments()

            loader.load_arguments(command)

            self.argument_registry.arguments.update(
                loader.argument_registry.arguments)
            self.extra_argument_registry.update(loader.extra_argument_registry)

            super(CliCommandsLoader, self).load_arguments(command)


# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class CliCommandGroup(knack.commands.CommandGroup):
    def command(self, name, handler_name, **kwargs):
        """
        Register a command into the command table

        :param name: The name of the command
        :type name: str
        :param handler_name: The name of the handler that will be  applied to
               the operations template
        :type handler_name: str
        :param kwargs: Kwargs to apply to the command. Possible values:
               `client_factory`, `arguments_loader`, `description_loader`,
               `description`,`formatter_class`, `table_transformer`,
               `deprecate_info`, `validator`, `confirmation`.
        """

        # Stash custom command output markers by command namespace into the
        # cli's invocation context.
        # Custom outputs found in command `output` can be:
        # output=fn|str=table,json,tsv|azdata.cli.core.layout.BaseLayout
        cmd = ' '.join(handler_name.split('_'))
        invocation = self.command_loader.cli_ctx.invocation
        custom = invocation.data.get('custom', {})
        custom[cmd] = kwargs.get('output')
        invocation.data['custom'] = custom

        super(CliCommandGroup, self).command(name, handler_name, **kwargs)
