# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from azdata.cli.core.logging import get_logger
from jinja2 import Template as JinjaTemplate
from kubernetes import config
from pathlib import Path, PureWindowsPath
from string import Template
from azdata.cli.core.constants import (IO_DELIM, KEY_VALUE_SPLIT, IS_WINDOWS)
from jsonpath_ng.ext import parse
from jsonpatch import JsonPatch
from azdata.cli.core.exceptions import CliError
from azdata.cli.core.prompt import (prompt, prompt_pass, prompt_for_input)
from azdata.cli.core.configuration import Configuration
from azdata.cli.core.constants import (AZDATA_USERNAME, AZDATA_PASSWORD)
from azdata.cli.core.util import FileUtil
from humanfriendly import (AutomaticSpinner)

import json
import platform
import os
import re
import signal
import sys
import time

logger = get_logger(__name__)

controller_env_list = [AZDATA_USERNAME,
                       AZDATA_PASSWORD]

active_directory_env_list = ["DOMAIN_SERVICE_ACCOUNT_USERNAME",
                             "DOMAIN_SERVICE_ACCOUNT_PASSWORD"]

# Retry parameters for connection erros
#
CONNECTION_RETRY_INTERVAL_SECONDS = 5
CONNECTION_RETRY_ATTEMPTS = 12


def display(msg):
    """
    Displays message to `stdout` and send the message to `logger.info()` for
    debugging. If the `--output` was included to indicate structured output,
    both `stdout` and `logger.info()` under the console-handler are buffered and
    flushed at the end and sent to the respective `--output <format>`.

    Note: This is only temporary (iter-1) as we begin to refactor and separate
    existing info logging statements under `cluster` that doubled as both
    UX/stdout messages and debug logging to the respective file-logger and
    console-logger handlers. For (iter-1), we replaced any `logger.info`
    statement occurrences with this function `util.display` which just wrappers
    `info.logger`.

    :param msg: The message to print to `stdout` and respective log handlers.
    """
    from azdata.cli.core.output import (OutputStream, OutputProducer)

    # -- send message to logger(s) --
    logger.info(msg)

    # -- send message to `stdout` --
    # suppress if --debug|--verbose  were include so we do not double log both
    # to the console-handler and to `stdout`
    cli_config = Configuration()
    if OutputProducer.structured_output or \
            (not cli_config.debug and not cli_config.verbose):
        OutputStream().stdout.write(msg)


def check_security_config(security_object):
    """
    Checks if security config is valid
    """
    if not security_object:
        raise ValueError("security object json given is empty")

    security_config_parameters = [
        "ouDistinguishedName",
        "domainDnsName",
        "clusterAdmins",
        "clusterUsers"
    ]

    # Check required configs not empty
    #
    for security_config in security_config_parameters:
        if security_config not in security_object:
            raise ValueError("Security configuration paramater '" + security_config + "' must be specified in Active Directory mode.")

    # Check clusterAdmins could only have one group allowed
    #
    if len(security_object["clusterAdmins"]) != 1:
        raise ValueError("Security configuration parameter 'clusterAdmins' must contain a single domain group. Values specified: " + ','.join(security_object["clusterAdmins"]))
    if len(security_object["clusterUsers"]) == 0:
        raise ValueError("Security configuration paramater 'clusterUsers' must be specified in Active Directory mode, currently empty.")

    # Check external ad mode
    #
    if "useInternalDomain" not in security_object or not security_object["useInternalDomain"]:
        if "dnsIpAddresses" not in security_object or not security_object["dnsIpAddresses"]:
            raise ValueError("Security configuration paramater 'dnsIpAddresses' must be specified in Active Directory mode.")

        if "domainControllerFullyQualifiedDns" not in security_object or not security_object["domainControllerFullyQualifiedDns"]:
            raise ValueError("Security configuration paramater 'domainControllerFullyQualifiedDns' must be specified in Active Directory mode.")

    return True


def security_check(stdout, force, config_object):
    """
    Checks the environment sets and the config for the security object
    :param stdout:
    :param force:
    :param config_object:
    :return:
    """
    try:
        if is_active_directory_security_enabled(config_object):
            if not force and sys.stdin.isatty():
                read_environment_variables('activeDirectory')
                stdout('')
            check_environment_variables('activeDirectory')

            # check security config for CU1
            # check the elements need to be in security config and clusterAdmins only has one element
            #
            if 'activeDirectory' in config_object['security']:
                check_security_config(config_object["security"]["activeDirectory"])
            else:
                check_security_config(config_object["security"])

    except Exception as e:
        raise ValueError('Failed to set security configuration. '
                         'Please check the deployment documentation https://aka.ms/bdc-deploy . {0}'.format(e))


def is_active_directory_security_enabled(config_object):
    """
    Returns if the active directory security is enabled by the config object. 
    Here are the two cases considered that active directory security has been configured: 
    1). The config object has the security.activeDirectory subobject. 
    2). The config object has the security object and the security object doesn't have other new functionality, 
        such as: security.privileged, security.allowNodeMetricsCollection and so on. 
    The 2nd case is added to support the backward compatiblity of the flat layout of active directory, which 
    is used on and before CU1 release cluster. 

    :param config_object:
    :return:
    """

    if 'security' not in config_object or not config_object['security']:
        return False

    if 'activeDirectory' in config_object['security']:
        return True

    # If user has the new functionality configured in the security object, we consider that the active directory
    # security is not configured in the object.
    #
    if 'privileged' in config_object['security'] or 'allowDumps' in config_object['security'] \
        or 'allowNodeMetricsCollection' in config_object['security'] or 'allowPodMetricsCollection' in config_object['security']:
        return False
    else:
        return True


def control_config_check(stdout, config_object):
    """
    Checks if config is valid
    :param stdout:
    :param config_object:
    :return:
    """
    if 'endpoints' in config_object['spec']:
        if is_active_directory_security_enabled(config_object):
            endpoints_no_dns = list(filter(lambda x: 'dnsName' not in x, config_object['spec']['endpoints']))
            if endpoints_no_dns:
                error_endpoints = map(lambda x: x['name'], endpoints_no_dns)
                raise ValueError('dnsName required for endpoint(s): {0}'.format(', '.join(error_endpoints)))

        for endpoint in config_object['spec']['endpoints']:
            if 'dnsName' in endpoint and not re.match(r"^[a-z]([-a-z0-9]*[a-z0-9])?(\.[a-z]([-a-z0-9]*[a-z0-9])?)+$", endpoint['dnsName']):
                raise ValueError('dnsName {0} for endpoint {1} is not a valid RFC-1035 DNS domain.'.format(endpoint['dnsName'], endpoint['name']))


def read_config(config_profile, config_filename):
    """
    Read selected profile into object
    :param config_profile:
    :param config_filename:
    :return:
    """
    try:
        config = os.path.join(config_profile, config_filename)
        config_string = substitute_env_variables(config)
        config_object = read_config_json_string(config_string)
        return config_object
    except Exception as e:
        raise ValueError('Failed to read config profile {0}'.format(e))


def check_missing(stdout, force, config_object, help_object, config_filename):
    """
    Check for missing values and prompt user for them or raise an error in notty
    :param stdout:
    :param force:
    :param config_object:
    :param help_object:
    :param config_filename:
    :return:
    """
    missing_values = parse_for_missing(config_object)
    if len(missing_values) > 0:
        if not force:
            stdout('\nNeeded Configuration Values in {0}'.format(config_filename))
            try:
                for key in missing_values:
                    expr = parse(key)
                    find_result = [match.value for match in expr.find(help_object)]
                    help_result = find_result[0] if len(find_result) > 0 else None
                    name = help_result['name'] if help_result else key
                    description = help_result['description'] if help_result else \
                        'Please consult documentation: ' \
                        'https://aka.ms/bdc-deploy'
                    stdout('\n{name}\n- Config Path: {key}\n- Description: {description}'
                           .format(name=name, key=key, description=description))
                    result = prompt_for_input('- Please provide a value: ', '', padding=False, strip=True)
                    expr.update(config_object, result)
            except Exception as e:
                raise ValueError('Failed to set needed configuration values, '
                                 'please initialize your chosen config profile '
                                 'with \'{0} bdc config init\' and edit manually. '
                                 'Missing Configuration Values: {1}'.format(Configuration().CLI_NAME, missing_values))
        else:
            raise ValueError('Missing Configuration Values: {0}'.
                             format(missing_values))


def read_environment_variables(target):
    """
    Reads environment variables for the given target
    :param client: client handle
    :param target: target to read environment variables
    :return:
    """
    for env in get_environment_list_by_target(target):
        msg = 'Please provide a value for {}:'.format(display_env_name(env))
        if not os.environ.get(env):
            if 'PASSWORD' in env:
                result = prompt_pass(msg)
            else:
                result = prompt(msg)
            os.environ[env] = result.strip()


def read_config_json_string(json_string):
    """
    Reads a json string into a python dictionary
    """
    try:
        config_object = json.loads(json_string)
        return config_object
    except Exception as e:
        raise ValueError("Invalid json string")


def get_config_from_template(template_file, cluster_object):
    """
    Get kubernetes config from template using environment variables.
    """
    with open(template_file, "r") as f:
        template = f.read()
        s = JinjaTemplate(template)
        config = s.render(model=cluster_object)
        return config


def substitute_env_variables(template_file):
    """
    Substitute env variables in a file and return the string.
    """
    with open(template_file, "r") as f:
        template = f.read()
        s = Template(template)
        config = s.substitute(os.environ)
        return config


def parse_for_missing(config_object):
    """
    Parses the config json object for missing/falsey values
    Returns a list of json paths for the missing values
    :param config_object:
    :return: missing:
    """
    try:
        missing = []

        def iterate(dictionary, root):
            for key, value in dictionary.items():
                new_root = root + '.' + key if root else key
                if isinstance(value, dict):
                    iterate(value, new_root)
                    continue
                elif isinstance(value, list):
                    for index, item in enumerate(value):
                        arr_root = new_root + '[' + str(index) + ']'
                        if isinstance(item, dict):
                            iterate(item, arr_root)
                        elif item == "":
                            missing.append(arr_root)
                    continue
                if value == "":
                    missing.append(new_root)

        iterate(config_object, '')
        return missing
    except Exception as e:
        raise ValueError(
            'Failed to parse for missing values in config file: {}'.format(e))


def set_env_vars(env_list):
    """
    Sets environment variables based on the given list
    :param env_list:
    :return:
    """
    try:
        for param in env_list.split(','):
            var = param.strip().split('=')
            os.environ[var[0]] = var[1]
    except Exception as e:
        raise ValueError(
            'Please pass a list of environment variables as a key/value list, '
            'i.e. var1=val1,var2=val2')


def display_env_name(env):
    """
    Provides a user friendly display name for environment variables
    :param env:
    :return:
    """
    return env.strip()


class TimeoutError(Exception):
    pass


def handle_timeout(signum, frame):
    """throws timeout error
    """
    raise TimeoutError("Timeout occurred")


def with_timeout(timeout, function_to_run, *args, **kwargs):
    """Runs a function with timeout
    Args:
        timeout: the number of seconds to wait before fail the function with
        timeout error
        function_to_run: the function to run with given timeout
    """
    signal_handler = signal.signal(signal.SIGALRM, handle_timeout)
    signal.alarm(timeout)
    try:
        function_to_run(*args, **kwargs)
    finally:
        signal.signal(signal.SIGALRM, signal_handler)
        signal.alarm(0)


def check_and_set_kubectl_context():
    """
    Check and set environment var for kubectl command context
    """
    try:
        if os.getenv("KUBERNETES_SERVICE_HOST"):
            config.load_incluster_config()
        elif os.getenv("KUBECTL_CONTEXT") is None:
            load_kube_config()
        else:
            load_kube_config(os.environ["KUBECTL_CONTEXT"])
    except Exception as e:
        logger.debug(e)
        logger.error("Failed to complete kube config setup.")
        raise Exception("Failed to complete kube config setup.")


def load_kube_config(context=None):
    #
    # Python Kubernetes Library does not handle cases where the KUBECONFIG
    # environment variable can have multiple
    # kubeconfig files listed. It's an open issue on github -
    # https://github.com/kubernetes-client/python-base/pull/94
    # This is a temporary workaround until this change gets checked in.
    #
    DEFAULT_KUBE_CONFIG_LOCATION = None
    FILE_SEPARATOR = None
    if platform.system() == 'Windows':
        DEFAULT_KUBE_CONFIG_LOCATION = str(
            PureWindowsPath(str(Path.home()) + '/.kube/config'))
        FILE_SEPARATOR = ';'
    else:
        DEFAULT_KUBE_CONFIG_LOCATION = str(Path.home()) + '/.kube/config'
        FILE_SEPARATOR = ':'
    kube_config_files = os.environ.get('KUBECONFIG')
    kube_context_loaded = False
    config_file = None

    if kube_config_files:
        config_file = None
        for config_file in kube_config_files.split(FILE_SEPARATOR):
            if config_file:
                # kubectl gives precedence to the order in which the files
                # are listed.
                #
                try:
                    config.load_kube_config(config_file=config_file,
                                            context=context)
                    kube_context_loaded = True
                    break
                except Exception as e:
                    logger.debug("Could not find context %s in file - %s" % (
                    context, config_file))
        if not kube_context_loaded:
            logger.error(
                "Failed to load context %s from the listed kubeconfig files : "
                "%s." % (
                context, kube_config_files.split(FILE_SEPARATOR)))
            raise Exception("Failed to load kubeconfig.")
        else:
            logger.debug("Loaded kube config from %s" % (config_file))
    else:
        logger.debug("Loading default kube config from %s " % (
        DEFAULT_KUBE_CONFIG_LOCATION))
        config.load_kube_config(context=context)

    # set kubectl context to current context if not set
    #
    if context is None:
        contexts, current_context = config.list_kube_config_contexts(
            config_file=config_file)
        if current_context:
            os.environ["KUBECTL_CONTEXT"] = current_context['name']
        else:
            logger.error("No active context is set in kubeconfig.")
            raise Exception("No active context is set in kubeconfig.")
    logger.debug("Kubectl context set to %s" % os.environ["KUBECTL_CONTEXT"])


def add_node_selector_to_controller_config(controller_yaml_obj, clusterLabel, nodeLabel):
    """
    Add node selector field to controller spec
    """
    if "nodeSelector" not in controller_yaml_obj["spec"]["template"]["spec"]:
        controller_yaml_obj["spec"]["template"]["spec"]["nodeSelector"] = {}

    if clusterLabel:
        controller_yaml_obj["spec"]["template"]["spec"]["nodeSelector"]["mssql-cluster"] = clusterLabel

    if nodeLabel:
        controller_yaml_obj["spec"]["template"]["spec"]["nodeSelector"]["mssql-resource"] = nodeLabel


def add_node_selector_to_controller_config_map(controller_config_map_yaml_obj, clusterLabel, nodeLabel):
    """
    Add node selector field to controller config map
    """
    configJsonStr = controller_config_map_yaml_obj["data"][
        "mssql-controller.conf"]

    # The value of key "mssql-controller.conf" is a json string.
    # We add label attribute and put the json string back.
    jsonObj = json.loads(configJsonStr)
    if clusterLabel:
        jsonObj["spec"]["clusterLabel"] = clusterLabel

    if nodeLabel:
        jsonObj["spec"]["nodeLabel"] = nodeLabel

    controller_config_map_yaml_obj["data"][
        "mssql-controller.conf"] = json.dumps(jsonObj)


def get_environment_list_by_target(target):
    """
    Gets the required environment variable list for the given target
    """
    env_list = []
    if target == "clean":
        env_list = controller_env_list
    elif target == "config":
        env_list = controller_env_list
    elif target == "controller":
        env_list = controller_env_list
    elif target == "cluster":
        env_list = controller_env_list
    elif target == "activeDirectory":
        env_list = active_directory_env_list
    elif target == "delete":
        env_list = controller_env_list
    elif target == "password":
        env_list = controller_env_list
    return env_list


def check_environment_variables(target):
    """
    Check if all necessary environment variables are set.
    """

    env_list = get_environment_list_by_target(target)

    missing_env = list()
    for env in env_list:
        if os.getenv(env) is None:
            missing_env.append(env)
    if len(missing_env) > 0:
        logger.error(
            "Please set the following environment variable(s): %s." %
            missing_env)
        sys.exit(1)


def log_error(error_message, error_detail):
    """
    Log error messages
    :param error_message : Error thrown
    :param error_detail : Error info
    """

    logger.error(
        error_message + " : " + error_detail + ". More detail can be found in "
                                               "the azdata.log file")


def retry(func, *func_args, **kwargs):
    """
    Function retry until max limit
    :param func : Function to retry
    :param func_args : Arguments to func
    :param int retry_count : Retry attempts
    :param int retry_delay : Retry delay
    :param string retry_method : func literal
    :param tuple retry_on_exceptions : Exceptions to ignore while retrying func
    """

    retry_count = kwargs.get("retry_count", CONNECTION_RETRY_ATTEMPTS)
    retry_delay = kwargs.get("retry_delay", CONNECTION_RETRY_INTERVAL_SECONDS)
    retry_method = kwargs.get("retry_method", None)
    retry_on_exceptions = kwargs.get("retry_on_exceptions", None)
    exception_caused = None
    for i in range(retry_count):
        try:
            return_value = func(*func_args)
            return return_value if return_value is not None else True
        except retry_on_exceptions as e:
            exception_caused = e
            logger.debug("Waiting for %d seconds before trying to %s again" % (
            retry_delay, retry_method))
            time.sleep(retry_delay)
            continue
    logger.debug(exception_caused)
    log_error(
        str(getattr(exception_caused.reason, '__context__', exception_caused)),
        "Failed to %s after retrying for %d minute(s)." % (
        retry_method, (retry_count * retry_delay) / 60))
    raise Exception("Failed to %s after retrying for %d minute(s)." % (
    retry_method, (retry_count * retry_delay) / 60))


# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class DeploymentConfigUtil(object):
    """
    Utility methods for handling deployment configuration files.
    """
    @staticmethod
    def get_config_map(config_dir):
        """
        Gets the available config profiles and returns them as a dictionary where
        the key/value are the config profile name
        :param config_dir:
        :return dictionary of config profiles with directory name being the key:
        """
        files = {}
        for f in os.listdir(config_dir):
            if f != 'help':
                files[f] = f
        return files

    @staticmethod
    def get_config_display_names(config_profile_dict):
        """
        Returns a display friendly list of config profile names
        :param config_profile_dict:
        :return:
        """
        display_list = list(config_profile_dict.keys())
        display_list.sort()
        return display_list

    @staticmethod
    def config_list(config_dir, config_profile=None, config_type='control'):
        """
        Lists available configuration file choices.
        :param config_dir:
        :param config_profile:
        :param config_type:
        :return:
        """
        try:
            configs = DeploymentConfigUtil.get_config_map(config_dir)
            if not config_profile:
                config_names = DeploymentConfigUtil.get_config_display_names(configs)
                return config_names
            elif config_profile in configs.keys():
                config_profile = os.path.join(config_dir, configs[config_profile])
                config_file = os.path.join(config_profile, config_type + '.json')

                if os.path.isfile(config_file):
                    file = open(config_file, 'r')
                    text = file.read()
                    return json.loads(text)
                else:
                    raise ValueError('Invalid config type: `{0}`'.format(config_type))
            else:
                raise ValueError('Invalid default config profile, please consult [bdc '
                                 'config list] for available types')
        except ValueError as e:
            raise CliError(e)
        except Exception as e:
            raise CliError(e)

    @staticmethod
    def save_config_profile(target, source, config_dir, config_files, config_map, force):
        """
        Saves the config profile from the source to the given target.
        :param target:
        :param source:
        :param config_dir:
        :param config_files:
        :param config_map:
        :param force:
        :return:
        """
        if not os.path.isdir(target):
            os.makedirs(target)

        result_path = os.path.join(config_dir, config_map[source.lower()])

        for config_filename in config_files:
            target_cluster_config_file = os.path.join(target, config_filename)
            if not force and os.path.isfile(target_cluster_config_file):
                raise FileExistsError('Please specify a directory that is empty '
                                      'or use the force flag: config init -t {0} --force'
                                      .format(target))

            # Read the file and parse into Python dictionary
            cluster_config_file = os.path.join(result_path, config_filename)
            profile = FileUtil.read_json(cluster_config_file)

            if not IS_WINDOWS:
                with AutomaticSpinner('Fetching {0} config'.format(source),
                                      show_time=True):
                    with open(target_cluster_config_file, 'w') as outfile:
                        json.dump(profile, outfile, indent=4)
            else:
                with open(target_cluster_config_file, 'w') as outfile:
                    json.dump(profile, outfile, indent=4)

        return target

    @staticmethod
    def write_config_file(filepath, config_object):
        """
        Writes the config object at the given filepath
        :param filepath:
        :param config_object:
        :return:
        """
        # -- write config --
        if not IS_WINDOWS:
            with AutomaticSpinner('Writing {0}'.format(filepath),
                                  show_time=True):
                with open(filepath, 'w') as outfile:
                    json.dump(config_object, outfile, indent=4)
        else:
            with open(filepath, 'w') as outfile:
                json.dump(config_object, outfile, indent=4)

    @staticmethod
    def config_add(config_file, json_values):
        """
        Add new key and value to the given config file
        :param config_file:
        :param json_values:
        :return:
        """
        config_object = FileUtil.read_json(config_file)
        DeploymentConfigUtil.inline_patch(config_object, json_values, 'add')
        return config_object

    @staticmethod
    def config_replace(config_file, json_values):
        """
        Replace the value of a given key in the given config file
        :param config_file:
        :param json_values:
        :return:
        """
        config_object = FileUtil.read_json(config_file)
        DeploymentConfigUtil.inline_patch(config_object, json_values, 'replace')
        return config_object

    @staticmethod
    def config_remove(config_file, json_path):
        """
        Remove a key from the given config file
        :param config_file:
        :param json_path:
        :return:
        """
        config_object = FileUtil.read_json(config_file)
        DeploymentConfigUtil.inline_patch(config_object, json_path, 'remove')
        return config_object

    @staticmethod
    def config_patch(config_file, patch_file):
        """
        Patch a given file against the given config file
        :param config_file:
        :param patch_file:
        :return:
        """
        config_object = FileUtil.read_json(config_file)

        if os.path.isfile(patch_file):
            patch_json = FileUtil.read_json(patch_file)
            patch = patch_json.get('patch')
            if patch:
                patches = JsonPatch(patch)
                for p in patches:
                    path = p.get('path')
                    if not path:
                        raise ValueError('Please specify a \'path\' for all patches')
                    op = p.get('op')
                    if not op:
                        raise ValueError('Please specify an \'op\' for all patches')
                    value = p.get('value')
                    if value is None and op != 'remove':
                        raise ValueError('Please specify a \'value\' for all patches other than \'remove\' ops')
                    DeploymentConfigUtil.patch_value(config_object, path, value, op)
            else:
                raise ValueError('The patch file must be json and start with a key of \'patch\' with its value '
                                 'as an array of patches.')
        else:
            raise ValueError('Please specify a file for the --patch-file argument.  The patch file must be json '
                             'and start with a key of \"patch\" with its value as an array of patches.')

        return config_object

    @staticmethod
    def inline_patch(config_object, json_values, op):
        """
        Values provided as an inline key=value comma separated list
        :param config_object:
        :param json_values:
        :param op:
        :return:
        """
        json_values = re.split(IO_DELIM, json_values)
        for param in json_values:
            var = re.split(KEY_VALUE_SPLIT, param.strip())
            json_path = var[0]
            value = None
            if len(var) > 1:
                if os.path.isfile(var[1].strip()):
                    value = FileUtil.read_json(var[1])
                else:
                    # Replace escaped commas and equality signs with normal symbols
                    # for the actual operation
                    value = var[1].replace('\\,', ',')
                    value = value.replace('\\=', '=')
                    try:
                        value = json.loads(value)
                    except Exception as e:
                        if value.startswith('{'):
                            raise ValueError(e)
            DeploymentConfigUtil.patch_value(config_object, json_path, value, op)

    @staticmethod
    def patch_value(config_object, json_path, value, op):
        """
        Patch the config_object with the value at the json_path with the given op (operation)
        :param config_object:
        :param json_path:
        :param value:
        :param op:
        :return:
        """
        msg = '\'{op}\' at \'{json_path}\' with value of \n{val}\n' \
            .format(op=op, json_path=json_path, val=json.dumps(value, indent=4))

        # If doing a replace and using conditional subsetting, use the json path library
        if op == 'replace' and ('$' in json_path or '@' in json_path):
            try:
                json_path = DeploymentConfigUtil.replace_path(json_path, ['/'], '.')
                expr = parse(json_path)
                find_result = [match.value for match in expr.find(config_object)]
                if len(find_result) > 0:
                    expr.update(config_object, value)
                else:
                    raise ValueError(
                        'Your given json path does not exist. Please give a json '
                        'path that exists.')
            except Exception as e:
                raise ValueError(
                    '{msg}\nDetails: {error}\n\nPlease consult: '
                    'https://jsonpath.com/'
                        .format(msg=msg, error=e))
        # Otherwise, use the patch library
        else:
            try:
                # Replace jsonpath related characters from the path with json patch syntax
                json_path = DeploymentConfigUtil.replace_path(json_path, ['.', '[', ']', '//'], '/')
                if json_path.startswith('$'):
                    json_path = json_path[1:]
                if not json_path.startswith('/'):
                    json_path = '/' + json_path
                if json_path.endswith('/'):
                    json_path = json_path[:-1]
                patch = JsonPatch([{'op': op, 'path': json_path, 'value': value}])
                patch.apply(config_object, True)
            except Exception as e:
                raise ValueError(
                    '{msg}\nDetails: {error}\n\nPlease consult: '
                    'http://jsonpatch.com/'
                        .format(msg=msg, error=e))

    @staticmethod
    def set_config_value(config_object, json_path, value, op=None):
        """
        Sets the given value at the json_path in the config_object
        :param config_object:
        :param json_path:
        :param value:
        :return:
        """
        if not op or op == 'replace':
            json_path = DeploymentConfigUtil.replace_path(json_path, ['/'], '.')
            # Attempt to update the path if it already exists
            expr = parse(json_path)
            find_result = [match.value for match in expr.find(config_object)]
            if len(find_result) > 0:
                expr.update(config_object, value)
                return
            elif '$' in json_path:
                raise ValueError(json_path)

        # The path did not exist, so now try to patch it
        json_path = DeploymentConfigUtil.replace_path(json_path, ['.', '[', ']', '//'], '/')
        if not json_path.startswith('/'):
            json_path = '/' + json_path
        if json_path.endswith('/'):
            json_path = json_path[:-1]
        if not op:
            try:
                # Attempt a replace
                patch = JsonPatch(
                    [{'op': 'replace', 'path': json_path, 'value': value}])
                patch.apply(config_object, True)
            except:
                # If replace did not work, attempt an add
                patch = JsonPatch(
                    [{'op': 'add', 'path': json_path, 'value': value}])
                patch.apply(config_object, True)
        else:
            patch = JsonPatch([{'op': op, 'path': json_path, 'value': value}])
            patch.apply(config_object, True)

    @staticmethod
    def replace_path(json_path, old_sep, new_sep):
        """
        Replaces jsonpath with slashes for jsonpatch library if given as periods
        :param json_path:
        :param old_sep:
        :param new_sep:
        :return:
        """
        new_path = json_path.strip()
        for sep in old_sep:
            new_path = new_path.replace(sep, new_sep)
        return new_path
