# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

"""This module contains custom CLI exceptions."""

from __future__ import absolute_import
from knack.util import CLIError as KnackCliError

HTTP_UNAUTHORIZED = 401
"""HTTP Unauthorized access is denied due to invalid credentials."""

OPERATION_TIMED_OUT = 60
""" Failed to establish a new connection: [Errno 60] Operation timed out."""

CONNECTION_REFUSED = 61
"""Failed to establish a new connection: [Errno 61]. Connection refused."""

DEFAULT_EXCEPTION_STATUS = 1
"""
Defines the default generic exception status code. Status codes are mostly
useful for process I/O errors and HTTP errors. All subclass exceptions 
thrown will adopt this value if one is not provided.
"""


class HTTPCodes(object):
    """
    Defines the HTTP status codes.
    NOTE: Add more when needed.
    """

    # -- Success --
    @property
    def ok(self):
        return 200

    @property
    def created(self):
        return 201

    # -- Client Errors --
    @property
    def bad_request(self):
        return 400

    @property
    def unauthorized(self):
        return 401

    @property
    def forbidden(self):
        return 403

    @property
    def not_found(self):
        return 404

    @property
    def method_not_allowed(self):
        return 405

    # -- Server Errors --
    @property
    def bad_gateway(self):
        return 502

    @property
    def service_unavailable(self):
        return 503

    @property
    def gateway_timeout(self):
        return 504

http_status_codes = HTTPCodes()  # export


http_codes = {
    # Client Error.
    400: ('bad_request', 'bad'),
    401: ('unauthorized',),
    403: ('forbidden',),
    404: ('not_found', '-o-'),
    # Server Error.
    500: ('internal_server_error', 'server_error', '/o\\', '✗'),
    501: ('not_implemented',),
    502: ('bad_gateway',),
    503: ('service_unavailable', 'unavailable'),
    504: ('gateway_timeout',),
}


class CliError(KnackCliError):
    """Common ancestor for CLI related exceptions."""

    def __init__(self, reason, status=DEFAULT_EXCEPTION_STATUS):
        reason = Exception(reason) if isinstance(reason, str) else reason
        super(CliError, self).__init__(reason)
        self.status = status
        self.reason = reason

    @property
    def success(self):
        return False

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        self._status = status

    @property
    def reason(self):
        return self._reason

    @reason.setter
    def reason(self, reason):
        self._reason = reason


class ConfigError(CliError):
    """All errors related to CLI configuration files."""
    pass


class ControllerError(CliError):
    """All errors related to Controller APIS."""
    pass


class ClusterLogError(CliError):
    """All errors related to cluster logs."""
    pass


class KerberosExchangeError(CliError):
    """
    Any client authentication errors that occurred while performing the GSSAPI
    negotiate steps.
    """
    pass


class ApiError(CliError):
    """HTTP errors while calling APIs against server controller-plane."""

    def __init__(self, status=None, reason=None, http_resp=None):
        import json

        if http_resp:
            status = http_resp.status
            reason = http_resp.reason
            self.headers = http_resp.getheaders()

            try:
                # HTTP errors can have this json structure in the body.
                # Here we clean it up for exception printing.
                body = json.loads(http_resp.data)
                self.body = ''

                if 'Link' in body:
                    self.body = '   Link: ' + body['Link']

                if 'Message' in body:
                    self.body += '\n   Message: ' + body['Message']

                if 'ExceptionType' in body:
                    self.body += '\n   Exception Type: ' + body['ExceptionType']

                if not self.body:
                    self.body = http_resp.data

            except (IOError, Exception):
                self.body = http_resp.data
        else:
            self.body = None
            self.headers = None

        super(ApiError, self).__init__(reason, status)

    def __str__(self):
        """
        Custom error messages for HTTP exception.
        """
        error_message = '({0})\n'\
                        'Reason: {1}\n'.format(self.status, self.reason)
        if self.headers:
            error_message += 'HTTP response headers: {0}\n'.format(self.headers)

        if self.body:
            error_message += 'HTTP response body: {0}\n'.format(self.body)

        return error_message
