from azdata.cli.core.logging import get_logger
import azdata.cli.core.deploy as util
import os
import random
import string
import unicodedata

logger = get_logger(__name__)

#
# Error codes for failed password validation
#
class PasswordErrorCodes:
    SUCCESS        = 0
    TOO_SHORT      = 1
    TOO_LONG       = 2
    NOT_COMPLEX    = 3
    ENCODING_ERROR = 4
    DECODING_ERROR = 5
    CONTROL_CHARS  = 6


class PasswordManager:

    def __init__(self):
        pass

    @staticmethod
    def generate_random_sql_linux_password():
        """
        Generates a random SQL Server password per the password requirements of SQL on Linux
        """
        rand = random.SystemRandom()
        password_length = 32
        sample_space = string.ascii_letters + string.digits + '!$#%'
        return ''.join(rand.choice(sample_space) for i in range(password_length))

    @staticmethod
    def check_password_validity(target):
        """Check validity of passwords passed through environment variables

        Args:
            target: target to check password validity for
        """
        env_list = util.get_environment_list_by_target(target)

        for password_var in env_list:
            if not password_var.endswith("PASSWORD"):
                continue

            if password_var not in os.environ:
                continue

            if password_var != 'DOCKER_PASSWORD':
                is_success = PasswordManager.validate_password_and_print_if_error(os.environ[password_var])
                if not is_success:
                    import sys
                    logger.error("Password validation failed for '%s'" % password_var)
                    sys.exit(1)

    @staticmethod
    def print_password_error_message(error_code):
        """Print an error message if password can't be validated.

         Args:
             error_code(int): Error code
        """

        if error_code == PasswordErrorCodes.TOO_SHORT:
            print("The specified password does not meet SQL Server password policy requirements because it is "
                    "too short. The password must be at least 8 characters")
        elif error_code == PasswordErrorCodes.TOO_LONG:
            print("The specified password does not meet SQL Server password policy requirements because it is "
                    "too long. The password cannot exceed 128 characters")
        elif error_code == PasswordErrorCodes.NOT_COMPLEX:
            print("The specified password does not meet SQL Server password policy requirements because it "
                          "is not complex enough. The password must be at least 8 characters long and contain characters "
                          "from three of the following four sets: uppercase letters, lowercase letters, numbers, "
                          "and symbols.")
        elif error_code == PasswordErrorCodes.ENCODING_ERROR:
            print("The specified password contains a character that cannot be encoded to UTF-8. "
                             "Try using a password with only ASCII characters.")
        elif error_code == PasswordErrorCodes.DECODING_ERROR:
            print("The specified password contains a character that cannot be decoded. "
                             "Try using a password with only ASCII characters.")
        elif error_code == PasswordErrorCodes.CONTROL_CHARS:
            print("The specified password contains an invalid character. Valid characters "
                                   "include uppercase letters, lowercase letters, numbers, symbols, "
                                   "punctuation marks, and unicode characters that are categorized as alphabetic "
                                   "but are not uppercase or lowercase.")

    @staticmethod
    def validate_password_and_print_if_error(password):
        """Validate a password and print an error if it's not formed correctly.

        Args:
             password(str): Password

        Returns:
             True if valid, False otherwise
        """

        password_validation_result = PasswordManager.is_valid_password(password)

        if password_validation_result != PasswordErrorCodes.SUCCESS:
            PasswordManager.print_password_error_message(password_validation_result)
            return False
        else:
            return True

    @staticmethod
    def is_valid_password(password):
        """Determines if a given password matches SQL Server policy requirements

        Args:
            password (str): a password encoded as described by sys.stdin.encoding

        Returns:
            Integer error code. 0 for success, [1-6] otherwise. Error codes are defined
            in the PasswordErrorCodes class.

        Notes:
            This function cannot print anything to stdout or it will interfere with
            the validate-password flag in mssql-conf.py
        """

        # Characters that Windows treats as "special characters" for its password validation.
        #
        SYMBOL_CHARACTERS = "(`~!@#$%^&*_-+=|\\{}[]:;\"'<>,.?)/"

        # Known bug:
        # This function will return the incorrect character count for multi-byte encodings
        # For example, a smiley face emoji, U+1F601 will cause this function to allow shorter
        # passwords. However, this matches the behavior of Windows password policy validation,
        # and will thus cause no compatability issues or information loss with SQL Server.
        #
        if len(password) < 8:
            return PasswordErrorCodes.TOO_SHORT
        if len(password) > 128:
            return PasswordErrorCodes.TOO_LONG

        # Count the number of occurences of each type of character
        #
        contains_upper = 0
        contains_lower = 0
        contains_numeric = 0
        contains_symbol = 0

        for unicode_char in password:
            # Gets the general category assigned to a given unicode character
            # If it does not fall into any of these categories (For example, Sm, which includes U+221A)
            # then it is treated as a character for length, but does not count toward any category.
            #
            # This does not perfectly match the way that Windows evaluates characters, but is a
            # proper subset of characters accepted by Windows
            #
            # If a character is a "control character" like escape, then we print an error.
            #
            char_category = unicodedata.category(unicode_char)
            if unicode_char in SYMBOL_CHARACTERS:
                contains_symbol = 1
            elif char_category == 'Ll':
                contains_lower = 1
            elif char_category == 'Lu':
                contains_upper = 1
            elif char_category == 'Nd':
                contains_numeric = 1
            elif char_category == 'Cc':
                return PasswordErrorCodes.CONTROL_CHARS

        if (contains_upper + contains_lower + contains_numeric + contains_symbol) >= 3:
            return PasswordErrorCodes.SUCCESS

        else:
            return PasswordErrorCodes.NOT_COMPLEX
