# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

import humanfriendly
import knack.prompting

# ------------------------------------------------------------------------ #
# Wrapper for prompts to check the AZDATA_OUTPUT_INTERACTIVE env var that gets set in the .azdata/config
# If the user is in an interactive env, they may set this env var to be 'no' in order to prevent interactivity
# from the cmd line.  If they do not set it, it will default to whatever their terminal supports.
# If they do set it as interactive and their terminal does not support interactivity, then it will still raise
# a NoTTY exception
# ------------------------------------------------------------------------- #


def interactive(func):
    """
    Decorator to wrap all promptable functions below
    :param func:
    :return:
    """
    def function_wrapper(*args, **kwargs):
        from azdata.cli.core.configuration import Configuration
        config = Configuration()
        if not config.interactive:
            raise knack.prompting.NoTTYException
        return func(*args, **kwargs)
    return function_wrapper


@interactive
def prompt_y_n(msg):
    """
    Prompts [y/n] to the user with the given message
    :param msg:
    :return:
    """
    return knack.prompting.prompt_y_n(msg)


@interactive
def prompt(msg):
    """
    Basic prompt to the user with the given message
    :param msg:
    :return:
    """
    return knack.prompting.prompt(msg)


@interactive
def prompt_pass(msg):
    """
    Prompts to the user with the given message and masks the input
    :param msg:
    :return:
    """
    return knack.prompting.prompt_pass(msg)


@interactive
def prompt_for_input(question, default=None, padding=True, strip=True):
    """
    More complex prompt to the user with default value and ability to control white space
    :param question:
    :param default:
    :param padding:
    :param strip:
    :return:
    """
    return humanfriendly.prompts.prompt_for_input(question, default, padding, strip)


@interactive
def prompt_for_choice(choices, default=None, padding=True):
    """
    Gives the user choices in a prompt
    :param choices:
    :param default:
    :param padding:
    :return:
    """
    return humanfriendly.prompt_for_choice(choices, default, padding)
