# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from azdata.cli.core.exceptions import CliError

import os


def raise_command_error(e, msg=None):
    """Manages the throwing of all CliError's."""
    from azdata.cli.core.configuration import Configuration
    from azdata.cli.core.exceptions import (
        CliError,
        HTTP_UNAUTHORIZED,
        CONNECTION_REFUSED,
        OPERATION_TIMED_OUT
    )

    status = 1
    cfg = Configuration()
    if isinstance(e, CliError):
        status = e.status

        if status == HTTP_UNAUTHORIZED:
            msg = 'You must be authenticated to run this command. First run ' \
                  'the {} login --help for more information on how ' \
                  'to login.'.format(cfg.CLI_NAME)
        elif status == CONNECTION_REFUSED or status == OPERATION_TIMED_OUT:
            msg = 'Cannot connect with endpoint. Verify that you defined the ' \
                  'endpoint correctly by running "{0} login ' \
                  '--endpoint <endpoint>" again. Also verify that the ' \
                  'endpoint is reachable.\n{1}'.\
                format(cfg.CLI_NAME, e)
        else:
            msg = 'Error processing command: "{0}"\n{1}'. \
                format(type(e).__name__, str(e.reason))

    if not msg:
        msg = 'Error processing command: "{0}"\n{1}'.\
            format(type(e).__name__, str(e))

    # -- log stacktrace only to the file-log --
    from azdata.cli.core.logging import CliLogging
    CliLogging.traceback(msg)

    raise CliError(msg, status=status)


def is_windows():
    """
    A common convenience function for windows environment check.
    :return: `True` if windows otherwise `False`.
    """
    return os.name == 'nt'


def merge(dict_1, dict_2):
    """
    Merge two dict's together.
    :return: The merged dict
    """

    result = dict_1.copy()
    result.update(dict_2)
    return result


def traceback():
    """
    Get the exection stacktrace.
    :return: The exception stacktrace.
    """
    import traceback
    import sys

    ex_t, ex_v, ex_tb = sys.exc_info()
    return traceback.format_exception(ex_t, ex_v, ex_tb)


def color_wrapper(color):
    import colorama

    def to_colored_msg(msg):
        return color + msg + colorama.Style.RESET_ALL

    return to_colored_msg


def env_to_section_option(env):
    """
    Takes an environment variable in conventional format `PREFIX_SECTION_OPTION`
    and parse of the `section` and `option` items.

    :param env: The environment variable in cli suggested format.
    :return: tuple of (section, option)
    """
    tokens = env.split('_')
    section = tokens[1]
    option = '_'.join(tokens[2:])
    return section, option


def minutes_elapsed(num_intervals, retry_interval):
    """
    Retrieve the number of minutes that have elapsed based on the number of
    intervals.
    """
    return num_intervals * retry_interval / 60


def basic_auth_token_to_username_password(token):
    """
    Convert the base64 authentication token into a decoded username and
    password.
    :param token: The base64 authentication token.
    :return: The decoded username and password as a `dict`.
    """
    try:
        from base64 import b64decode
        username_password = b64decode(token).decode('utf-8')
        token = username_password.split(':')

        if len(token) == 2:
            token = {'username': token[0], 'password': token[1]}
        else:
            raise ValueError('Unknown decoded token format.')

        return token
    except Exception as e:
        msg = 'Error decoding auth token {0}\n{1}'.format(type(e).__name__,
                                                          str(e))
        raise ValueError(msg)


def get_basic_auth_token(username, password, strip_prefix=True):
    """Gets HTTP basic authentication header (string).

    :return: The token for basic HTTP authentication.
    """
    import urllib3

    token = urllib3.util.make_headers(
        basic_auth=username + ':' + password
    ).get('authorization')

    # remove "Basic" from token format "Basic b64-encoded=="
    if strip_prefix:
        token = token.split()[1]

    return token


def eula_prompt(
        client,
        accept_eula=None,
        arg_allowed=False,
        force=False,
        bdc_eula=False):
    """
    Checks the given accept_eula param
    If no accept_eula param is given, checks for an env variable ACCEPT_EULA
    If no env var exists, prompts the user, if in an interactive environment
    :param client: The client of the calling command.
    :param accept_eula: whether the EULA has been accepted
    :param arg_allowed: If the command has --accept-eula as an arg, set this
                        to true for improved error messaging.
    :param force: If force, and eula is not accepted, method will raise error
    :param bdc_eula: Whether the BDC EULA is being checked
    :return: T`he str` "yes" for acceptance otherwise error.
    """
    from azdata.cli.core.configuration import Configuration
    from azdata.cli.core.prompt import prompt_y_n
    from knack.prompting import NoTTYException

    stdout = client.stdout
    cfg = Configuration()

    err_msg = 'Please accept the license terms to use this product through ' \
              'the prompt in\nan interactive env, or through the ' \
              'environment variable ACCEPT_EULA set \nto \'yes\' in a ' \
              'non-interactive env.'
    if arg_allowed:
        err_msg += ' For this command, you may also use the \narg ' \
                   '--accept-eula set to \'yes\'.'
    if force:
        err_msg += ' When using --force, please use the argument or' \
                   ' environment variable.'

    # Privacy Statement
    if bdc_eula or not client.profile.accept_eula:
        stdout('The privacy statement can be viewed at:')
        stdout('https://go.microsoft.com/fwlink/?LinkId=853010')
        stdout('')

    # SQL Server BDC EULA workflow
    bdc_accept_eula = accept_eula
    if bdc_eula:
        stdout('The license terms for SQL Server Big Data Cluster can be '
               'viewed at:')
        stdout('Enterprise: https://go.microsoft.com/fwlink/?linkid=2104292')
        stdout('Standard: https://go.microsoft.com/fwlink/?linkid=2104294')
        stdout('Developer: https://go.microsoft.com/fwlink/?linkid=2104079')
        stdout('')

        if not bdc_accept_eula:
            bdc_accept_eula = os.environ.get('ACCEPT_EULA', None)

        if not bdc_accept_eula and force:
            raise CliError(err_msg)

    # azdata EULA workflow
    # profile EULA is only used with azdata
    if not accept_eula:
        accept_eula = client.profile.accept_eula

    if not client.profile.accept_eula:
        stdout('The license terms for azdata can be viewed at:')
        stdout('https://aka.ms/eula-azdata-en')
        stdout('')

    # If no acceptance is provided, check the environment variable
    if not accept_eula:
        accept_eula = os.environ.get('ACCEPT_EULA', None)

    if not accept_eula and force:
        raise CliError(err_msg)

    # If the env var is not set, prompt the user
    # prompt is used for both eula's
    if not check_yes(accept_eula) or \
            (bdc_eula and not check_yes(bdc_accept_eula)):
        try:
            if cfg.interactive:
                yes = prompt_y_n('Do you accept the license terms?')
            else:
                raise NoTTYException
        except NoTTYException:
            raise CliError(err_msg)
        yes = 'yes' if yes else 'no'
    else:
        yes = accept_eula  # yes|no for no-tty

    # Raise error
    if not check_yes(yes):
        raise CliError(err_msg)
    return yes


def check_yes(accept):
    """
    Helper to determine if the eula was accepted
    :param accept:
    :return:
    """
    if not accept:
        return False
    elif accept.lower() != 'yes' and accept.lower() != 'y':
        return False
    else:
        return True


def load_kube_config(context=None):
    """
    Loads the kube config.
    :param context:
    :return:
    """
    from pathlib import Path, PureWindowsPath
    from kubernetes import config
    import platform

    # This is a temporary workaround until this change gets checked in.
    # Python Kubernetes Library does not handle cases where the KUBECONFIG
    # environment variable can have multiple kubeconfig files listed. It's an
    # open issue on github:
    # https://github.com/kubernetes-client/python-base/pull/94
    if platform.system() == 'Windows':
        DEFAULT_KUBE_CONFIG_LOCATION = str(PureWindowsPath(str(Path.home()) +
                                                           '/.kube/config'))
        FILE_SEPARATOR = ';'
    else:
        DEFAULT_KUBE_CONFIG_LOCATION = str(Path.home()) + '/.kube/config'
        FILE_SEPARATOR = ':'
    kube_config_files = os.environ.get('KUBECONFIG')
    kube_context_loaded = False
    config_file = None

    if kube_config_files:
        config_file = None
        for config_file in kube_config_files.split(FILE_SEPARATOR):
            if config_file:
                # kubectl gives precedence to the order in which the files are
                # listed.
                try:
                    config.load_kube_config(config_file=config_file,
                                            context=context)
                    kube_context_loaded = True
                    break
                except Exception as e:
                    raise("Could not find context %s in file - %s" %
                          (context, config_file))
        if not kube_context_loaded:
            raise("Failed to load context %s from the listed kubeconfig files:"
                  " %s." % (context, kube_config_files.split(FILE_SEPARATOR)))
    else:
        config.load_kube_config(context=context)

    # set kubectl context to current context if not set
    #
    if context is None:
        contexts, current_context = config.list_kube_config_contexts(
            config_file=config_file)
        if current_context:
            os.environ["KUBECTL_CONTEXT"] = current_context['name']
        else:
            raise Exception("No active context is set in kubeconfig.")


def get_controller_endpoint(clusterName, force_ip=False):
    """
    Get the controller endpoint
    :param clusterName:
    :param force_ip: in the case of a DNS, force return the endpoint with the ip address instead of the dns
    :return:
    """
    from kubernetes import client as k8sClient

    try:
        from azdata.cli.core.logging import get_logger
        logger = get_logger(__name__)

        host = ''
        port = ''
        service = k8sClient.CoreV1Api().read_namespaced_service(name='controller-svc-external', namespace=clusterName)
        if service.spec.type == 'LoadBalancer':
            ingress = service.status.load_balancer.ingress
            if ingress is not None and len(ingress) > 0:
                host = ingress[0].ip
                port = str(service.spec.ports[0].port)
        elif service.spec.type == 'NodePort':
            # Get the IP of any pod in the controller replica set for NodePort
            # service
            controller_pods = k8sClient.CoreV1Api().list_namespaced_pod(
                namespace=clusterName, label_selector='app=controller').items
            if controller_pods and len(controller_pods) > 0:
                host = controller_pods[0].status.host_ip
                port = str(service.spec.ports[0].node_port)
        else:
            raise Exception('Service type %s for service controller-svc-external is not supported.' % service.spec.type)

        if service.metadata.labels.get('dnsName', None) and not force_ip:
            host = service.metadata.labels['dnsName']

        return 'https://' + host + ':' + port

    except Exception as e:
        logger.error(e)
        raise Exception('Error when getting controller endpoint with kube config {0}'.format(e))


# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class FileUtil(object):
    @staticmethod
    def read_json(file_path):
        """
        Deserialize file containing a JSON document into a Python dict.
        :param file_path: The path to the JSON document.
        :return: A `dict` of JSON otherwise `None` if file not found.
        """
        from codecs import open
        import json

        content = None
        # for encoding in ['utf-8-sig', 'utf-8', 'utf-16', 'utf-16le',
        # 'utf-16be']:
        for encoding in ['utf-8']:
            try:
                with open(file_path, encoding=encoding) as f:
                    content = f.read()
            except (UnicodeError, UnicodeDecodeError):
                pass

        if not content:
            return None
        else:
            return json.loads(content)

    @staticmethod
    def read_text_file(file_path):
        """Helper to load text from a given file."""

        with open(FileUtil.strip_path(file_path), 'r') as file:
            data = file.read()

        return data

    @staticmethod
    def is_lnk_path(file_path):
        """
        Helper to identify if file path contains the special lnk `@` prefix.
        """
        return file_path.strip().startswith('@')

    @staticmethod
    def strip_path(file_path):
        """
        Helper to tidy up a file path and strip of any special lnk `@` prefix.
        """
        return file_path.strip().lstrip('@')

    @staticmethod
    def get_dict_from_csv(file_path):
        import csv

        with open(FileUtil.strip_path(file_path), 'r') as f:
            reader = csv.reader(f)
            headers = next(reader)
            column = {h: [] for h in headers}
            for row in reader:
                for h, v in zip(headers, row):
                    column[h].append(v)
            return column

# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


def singleton(cls, *args, **kw):
    instances = {}

    def _singleton():
        if cls not in instances:
            instances[cls] = cls(*args, **kw)
        return instances[cls]
    return _singleton


